﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Analysis_Part
{
    /// <summary>
    /// A class to represent the field of complex numbers, as well as all neccessary operations to run Algorithm B on it.
    /// </summary>
    [Serializable]
    public class Complex: Semiring
    {
        /// <summary>
        /// The real value of the complex number.
        /// </summary>
        private double real;
        /// <summary>
        /// The imaginary part of the complex number.
        /// </summary>
        private double imaginary;

        public Complex()
        {
          
        }

        /// <summary>
        /// Creates a new complex number based on another complex number
        /// this constructor is needed for Direct Product - constructor
        /// </summary>
        /// <param name="x"></param>
        public Complex(Complex x)
        {
            this.real = x.real;
            this.imaginary = x.imaginary;
        }

        /// <summary>
        /// Creates a new complex number with provided real and imaginary parts.
        /// </summary>
        /// <param name="real">The real part of the complex number.</param>
        /// <param name="imaginary">The imaginary part of the complex number.</param>
        public Complex(double real, double imaginary)
        {
            this.real = real;
            this.imaginary = imaginary;
        }
        /// <summary>
        /// Creates a complex number that also is a real number, i.e. with imaginary part 0.
        /// </summary>
        /// <param name="real">The (real) value of the number.</param>
        public Complex(double real)
        {
            this.real = real;
            this.imaginary = 0;
        }

        /// <summary>
        /// Generates a complex number from a string. The "inverse" of ToString.
        /// </summary>
        /// <param name="x"></param>
        public Complex(string x)
        {
            real = imaginary = 0;
            if (x.Length == 0)
            {
                real = 0;
                imaginary = 0;
                return;
            }
            char realsignum = '+';
            if (x[0] == '-')
            {
                realsignum = '-';
                x = x.Substring(1, x.Length - 1);
            }
            char imaginarysignum = '+';
            if(x.IndexOf('-')>-1) imaginarysignum = '-';

            string[] realdouble = x.Split(new char[]{imaginarysignum, 'i'});
            if (realdouble.Length == 1)
            {
                if(x.IndexOf('i') ==-1)
                    real = Convert.ToDouble(realsignum + realdouble[0]);
                else
                    imaginary = Convert.ToDouble(realsignum + realdouble[0]);
                return;
            }
            real = Convert.ToDouble(realsignum + realdouble[0]);
            imaginary = Convert.ToDouble(imaginarysignum+realdouble[1]);
        }
        /// <summary>
        /// Adds two complex numbers.
        /// </summary>
        /// <param name="c1">The first summand.</param>
        /// <param name="c2">The second summand.</param>
        /// <returns>The sum of c1 and c2.</returns>
        public static Complex operator+(Complex c1, Complex c2)
        {
            return new Complex(c1.real + c2.real, c1.imaginary + c2.imaginary);
        }
        /// <summary>
        /// Subtracts one complex number from another.
        /// </summary>
        /// <param name="c1">The minuend. </param>
        /// <param name="c2">The subtrahend.</param>
        /// <returns>c1-c2</returns>
        public static Complex operator -(Complex c1, Complex c2)
        {
            return new Complex(c1.real - c2.real, c1.imaginary - c2.imaginary);
        }
 
        /// <summary>
        /// Multiplies two complex numbers.
        /// </summary>
        /// <param name="c1">The first factor.</param>
        /// <param name="c2">The second factor.</param>
        /// <returns>The product of c1 and c2.</returns>
        public static Complex operator *(Complex c1, Complex c2)
        {
            return new Complex(c1.real * c2.real-c1.imaginary*c2.imaginary, c1.imaginary*c2.real + c2.imaginary*c1.real);
        }
        /// <summary>
        /// Converts a complex number into a string.
        /// </summary>
        /// <returns>A string representation of the complex number.</returns>
        public override string ToString()
        {
            if ((imaginary > 0) && (real != 0)) return (String.Format("{0}+{1}i", real, imaginary));
            if ((imaginary < 0) && (real != 0)) return (String.Format("{0}{1}i", real, imaginary));
            if (real == 0) return imaginary.ToString();
            else return real.ToString();
        }
        /// <summary>
        /// The one of the field of complex numbers.
        /// </summary>
        /// <returns>The complex number 1+0i.</returns>
        public static Complex One()
        {
            return new Complex(1, 0);
        }
        /// <summary>
        /// The zero of the field of complex numbers.
        /// </summary>
        /// <returns>The complex number 0+0i</returns>
        public static Complex Zero()
        {
            return new Complex(0, 0);
        }

        /// <summary>
        /// Adds 1 to a complex number.
        /// </summary>
        /// <param name="x">The number to be increased by one.</param>
        /// <returns>x+1</returns>
        public static Complex operator++(Complex x)
        {
            return new Complex(x.real + 1, x.imaginary);
        }
        /// <summary>
        /// Yields the additive inverse of a given complex number.
        /// </summary>
        /// <param name="x">The number for which the additive inverse is seeked.</param>
        /// <returns>-x</returns>
        public static Complex operator-(Complex x)
        {
            return new Complex(-x.real, -x.imaginary);
        }
        /// <summary>
        /// Divides two complex numbers.
        /// </summary>
        /// <param name="c1">The dividend.</param>
        /// <param name="c2">The divisor.</param>
        /// <returns>The quotient of c1 and c2.</returns>
        public static Complex operator /(Complex c1, Complex c2)
        {
            return new Complex((c1.real*c2.real+c1.imaginary*c2.imaginary)/(c2.real*c2.real+c2.imaginary*c2.imaginary), (c1.imaginary*c2.real-c1.real*c2.imaginary)/(c2.real*c2.real+c2.imaginary*c2.imaginary));
        }
        /// <summary>
        /// Checks whether two complex numbers are different.
        /// </summary>
        /// <param name="c1">The first number to be compared.</param>
        /// <param name="c2">The second number to be compared.</param>
        /// <returns>False, if both numbers are the same, true otherwise.</returns>
        public static bool operator !=(Complex c1, Complex c2)
        {
            return ((c1.real != c2.real) || (c1.imaginary != c2.imaginary));
        }
        /// <summary>
        /// Checks whether two complex numbers are equal.
        /// </summary>
        /// <param name="c1">The first number to be compared.</param>
        /// <param name="c2">The second number to be compared.</param>
        /// <returns>True, if both numbers are the same, false otherwise.</returns>
        public static bool operator==(Complex c1, Complex c2)
        {
            return ((c1.real == c2.real) && (c1.imaginary == c2.imaginary));
        }
        /// <summary>
        /// Divides each element in a complex array by the same number. Attention: This changes the vector, so this function has a side-effect!
        /// </summary>
        /// <param name="a">The array whose entries are to be divided. Attention: This function changes a!</param>
        /// <param name="b">The devisor.</param>
        public static void DivideArray(Complex[] a, Complex b)
        {
            for (int i = 0; i < a.Length; ++i) a[i] = a[i] / b;
        }
        /// <summary>
        /// Mulitplies each element in a complex array with the same number. Attention: This changes the vector, so this function has a side-effect!
        /// </summary>
        /// <param name="a">The array whose entries are to be multiplied. Attention: This function changes a!</param>
        public static void MultiplyArray(Complex[] a, Complex b)
        {
            for (int i = 0; i < a.Length; ++i) a[i] = a[i] * b;
        }
        /// <summary>
        /// Builds a deep copy of a givven complex-array.
        /// </summary>
        /// <param name="a">The array to be copied.</param>
        /// <returns>A deep copy of the array a.</returns>
        public static Complex[] CopyArray(Complex[] a)
        {
            Complex[] result = new Complex[a.Length];
            for (int i = 0; i < a.Length; ++i) result[i] = new Complex(a[i].real, a[i].imaginary);
            return result;
        }
        /// <summary>
        /// Copies a complex array and multiplies the copy with a given complex number.
        /// </summary>
        /// <param name="a">The array to be copied and multiplied.</param>
        /// <param name="b">The multiplicand.</param>
        /// <returns>A copy of a multiplied by b.</returns>
        public static Complex[] MultiplyCopyArray(Complex[] a, Complex b)
        {
            Complex[] result = CopyArray(a);
            MultiplyArray(result, b);
            return result;
        }
        /// <summary>
        /// Subtracts two complex arrays element-wise.
        /// </summary>
        /// <param name="a">The minuend array. Attention: The result is saved in a, so this has a side-effect on a!</param>
        /// <param name="b">The subtrahend array.</param>
        public static void SubtractArray(Complex[] a, Complex[] b)
        {
            for (int i = 0; i < a.Length; ++i) a[i] = a[i] - b[i];
        }
        /// <summary>
        /// Adds two complex arrays element-wise.
        /// </summary>
        /// <param name="a">The first summand array. Attention: The result is saved in a, so this has a side-effect on a!</param>
        /// <param name="b">The second summand array.</param>
        public static void AddArray(Complex[] a, Complex[] b)
        {
            for (int i = 0; i < a.Length; ++i) a[i] = a[i] + b[i];
        }
        /// <summary>
        /// Subtracts a complex number from a complex array element-wise.
        /// </summary>
        /// <param name="a">The minuend array. Attention: The result is saved in a, so this has a side-effect on a!</param>
        /// <param name="b">The subtrahend.</param>
        public static void SubtractArray(Complex[] a, Complex b)
        {
            for (int i = 0; i < a.Length; ++i) a[i] = a[i] - b;
        }
        /// <summary>
        /// Adds a complex number from a complex array element-wise.
        /// </summary>
        /// <param name="a">The first summand array. Attention: The result is saved in a, so this has a side-effect on a!</param>
        /// <param name="b">The second summand.</param>
        public static void AddArray(Complex[] a, Complex b)
        {
            for (int i = 0; i < a.Length; ++i) a[i] = a[i] + b;
        }
        
        

        /// <summary>
        /// Performs a downward elimination (Gaussian elimination) on a list of complex arrays interpreted as a matrix and an array interpreted as the result vector b in Ax=b.
        /// </summary>
        /// <param name="set">The list of row vectors (i.e. complex arrays) on which the downward elimination is to be performed. The function changes the contents of set, so it has a side effect on set.</param>
        /// <param name="vec">The result vector b in Ax=b.</param>
        /// <param name="p">A permutation, used to remember which columns got swapped in the process.</param>
        public static void DownwardElimination(List<Complex[]> set, Complex[] vec, Permutation p)
        {
            if (set.Count() == 0) return;
            for (int i = 0; (i< set[0].Length) && (i<set.Count()); ++i)
            {
                //Console.WriteLine("i: " + i.ToString() + ", vec.Length: " + vec.Length.ToString() + ", set[i].Length: " + set[i].Length.ToString());
                if (set[i][i] != Zero())
                {
                    vec[i] /= set[i][i];
                    DivideArray(set[i], set[i][i]);
                    for (int j = i+1; j < set.Count(); ++j)
                    {
                        vec[j] -= vec[i] * set[j][i];
                        SubtractArray(set[j], MultiplyCopyArray(set[i], set[j][i]));
                    }
                }
                else
                {
                    int k = i;
                    int j = i;
                    for (j = i; (j < set.Count()) && (set[j][k] == Zero()); ++j) 
                    {
                        for (k = i; (k < set[0].Length) && (set[j][k] == Zero()); ++k) ;
                        if (k == set[0].Length) k = i;
                        if (set[j][k] != Zero()) break;
                    }

                    //Console.WriteLine("In this matrix:");
                    //Console.WriteLine(new Matrix<Complex>(set));
                    //Console.WriteLine("I need to swap rows " + i + " and " + j + " and rows " + i + " and " + k);
                    if ((j < set.Count())&&(set[j][k] != Zero()))
                    {
                       // Console.WriteLine("swapping...");
                            swapListIndex(set, i, j);
                            swapArrayIndex(vec, i, j);
                            swapArrayIndexInList(set, i, k);
                            p.AddIndexPair(i--, k); // i needs to be explored again in the next step, therefore --.
                    }
                    //Console.WriteLine("After swapping: ");
                    //Console.WriteLine(new Matrix<Complex>(set));
                    
                }
            }
        }
        /// <summary>
        /// Performs an upward elimination (Gaussian elimination) on a list of complex arrays interpreted as a matrix and an array interpreted as the result vector b in Ax=b.
        /// </summary>
        /// <param name="set">The list of row vectors (i.e. complex arrays) on which the upward elimination is to be performed. The function changes the contents of set, so it has a side effect on set.</param>
        /// <param name="vec">The result vector b in Ax=b.</param>
        /// <returns>The linear combination x of vectors in set that forms vec, provided it exists. The boolean value in the result denotes if such a linear combination exists.</returns>
        public static LinearCombination<Complex> UpwardElimination(List<Complex[]> set, Complex[] vec)
        {
            LinearCombination<Complex> lc = new LinearCombination<Complex>();
            lc.CombinationExists = true;
            if (set.Count() == 0)
            {
                lc.CombinationExists = false;
                return lc;
            }
            lc.Multiplicands = new Complex[set[0].Count()];

            /*Console.WriteLine("Intermediate result before UpwardElimination: ");
            string debugausgabe = "set: ";
            for (int i = 0; i < set.Count(); ++i){ 
                for (int j = 0; j < set[i].Length; ++j) debugausgabe += set[i][j].ToString() + ", ";
                debugausgabe += "; ";
                }
            debugausgabe += '\n' + "vec: ";
            for (int i = 0; i < vec.Length; ++i) debugausgabe += vec[i].ToString() + ", ";
                Console.WriteLine(debugausgabe);*/
            
            for (int i = min(set[0].Length, vec.Length) - 1; i > -1; --i)
            {
                if (vec[i] != Zero() && set[i][i] == Zero()) lc.CombinationExists = false;
                for (int j = i-1; j > -1; --j)
                {
                    vec[j] -= vec[i] * set[j][i];
                    SubtractArray(set[j], MultiplyCopyArray(set[i], set[j][i]));
                }
                lc.Multiplicands[i]=vec[i];
            }
            for (int i = set[0].Length; i < vec.Length; ++i) if (vec[i] != Zero()) lc.CombinationExists = false;
            for (int i = vec.Length; i < set[0].Length; ++i) lc.Multiplicands[i] = Zero();

            return lc;
        }

        /// <summary>
        /// Interprets a list of complex-valued arrays as a matrix and transposes them.
        /// </summary>
        /// <param name="set">The list of arrays to be transposed. It is assumed that each array in list has the same length.</param>
        /// <returns>A list of complex arrays where the ith element in the jth array is the same as the jth element in the ith array in set.</returns>
        private static List<Complex[]> Transpose(List<Complex[]> set)
        {
            return Transpose<Complex>(set);
        }

        /// <summary>
        /// Makes a deep copy of a list of complex rows.
        /// </summary>
        /// <param name="set">The list that has to be copied.</param>
        /// <returns>A copy of the list set.</returns>
        public static List<Complex[]> DeepRowListCopy(List<Complex[]> set)
        {
            List<Complex[]> result = new List<Complex[]>();
            for (int i = 0; i < set.Count(); ++i)
            {
                Complex[] row = new Complex[set[i].Length];
                for (int j = 0; j < set[i].Length; ++j)
                {
                    row[j] = new Complex(set[i][j].real, set[i][j].imaginary);
                }
                result.Add(row);
            }
            return result;
        }
        /// <summary>
        /// Makes a deep copy of a complex array.
        /// </summary>
        /// <param name="x">The array to be copied.</param>
        /// <returns>A deep copy of the array x.</returns>
        public static Complex[] DeepArrayCopy(Complex[] x)
        {
            Complex[] result = new Complex[x.Length];
            for (int i = 0; i < x.Length; ++i) result[i] = new Complex(x[i].real, x[i].imaginary);
            return result;
        }
        /// <summary>
        /// Computes whether a vector can be found as a linear combination of a given set of vectors.
        /// </summary>
        /// <param name="set">A list of the vectors that may be linear combined to form vec.</param>
        /// <param name="vec">The vector that may be checked for being a linear combination of given set of vectors</param>
        /// <returns>A LinearCombination object that indicates whether the linear combination exists and, provided it exists, the factors.</returns>
        public static LinearCombination<Complex> LinearCombination(List<Complex[]> set, Complex[] vec) 
        {
            
            /*string debugoutput = "Check, whether (";
            for (int i = 0; i < vec.Length; ++i) debugoutput += vec[i].ToString() + (i<vec.Length-1?", ":"");
            debugoutput += ") can be found as linear combination of {";
            for (int i = 0; i < set.Count(); ++i)
            {
                debugoutput+="(";
                for (int j = 0; j < set[i].Length; ++j) debugoutput += set[i][j].ToString() + (j < set[i].Length - 1 ? ", " : "");
                if (i < set.Count() - 1) debugoutput += "), "; else debugoutput += ")}";
            }
            Console.WriteLine(debugoutput);*/
            
            
            List<Complex[]> setT = Transpose(DeepRowListCopy(set));
            Permutation p = new Permutation();
            Complex[] vecCopy = DeepArrayCopy(vec);
            //Console.WriteLine("Vor Downward-Elimination:");
            //Console.WriteLine(new Matrix<Complex>(setT));
            DownwardElimination(setT, vecCopy, p);
            //Console.WriteLine("Nach Downward-Elimination:");
            //Console.WriteLine(new Matrix<Complex>(setT));
            //Console.ReadLine();
            LinearCombination<Complex> lc = UpwardElimination(setT, vecCopy);
            p.ApplyTo<Complex>(lc.Multiplicands);

            
            /*debugoutput = " The answer is " + (lc.CombinationExists ? "yes, multiplicands are: " : "no.");
            if (lc.CombinationExists)
            {
                for (int i = 0; i < lc.Multiplicands.Length; ++i) debugoutput += lc.Multiplicands[i].ToString() + (i < lc.Multiplicands.Length - 1 ? ", " : "");
            }
            Console.WriteLine(debugoutput);*/
            
            
            return lc;
        }

        public static  bool allowedStrings(String s)
        {
            double imaginary;
            double real;
            real = imaginary = 0;
            if (s.Length == 0)
            {
                real = 0;
                imaginary = 0;
                return true;
            }
            char realsignum = '+';
            if (s[0] == '-')
            {
                realsignum = '-';
                s = s.Substring(1, s.Length - 1);
            }
            char imaginarysignum = '+';
            if (s.IndexOf('-') > -1) imaginarysignum = '-';

            string[] realdouble = s.Split(new char[] { imaginarysignum, 'i' });
            if (realdouble.Length == 1)
            {
                try
                {
                    if (s.IndexOf('i') == -1)
                        real = Convert.ToDouble(realsignum + realdouble[0]);
                    else
                        imaginary = Convert.ToDouble(realsignum + realdouble[0]);
                    return true;

                }
                catch (Exception) { return false; }
            
            }


            try
            {
                real = Convert.ToDouble(realsignum + realdouble[0]);
                imaginary = Convert.ToDouble(imaginarysignum + realdouble[1]);
                return true;
            }
            catch (Exception) { return false; }

        }

        public static string inputSchema()
        {
            return "3.4i-2 emptyspace next 3.567i0.0 ...\nPress Enter for next row";
        }

    }
}
