﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Analysis_Part
{
    /// <summary>
    /// The l-monoid for the lattice ([0,1], max, min) using the multiplication *, i.e. the usual multiplication in R.
    /// </summary>
    class LMonoidMaxTimes
    {
        /// <summary>
        /// The value of the LMonoidMaxTimes-object.
        /// </summary>
        uint z;
        const uint maxvalue = 1000000000;

        static Random random = new Random();


        public LMonoidMaxTimes()
        {

        }

        /// <summary>
        /// A copy constructor, creating a new LMonoidMaxTimes object with the same value as the provided one
        /// </summary>
        /// <param name="x">The LMonoidMaxTimes object to copy.</param>
         public LMonoidMaxTimes(LMonoidMaxTimes x)
        {
            z=x.z;
        }

         public LMonoidMaxTimes(String x)
         {
             if(x.Length<2)
             {
                 z=0; 
                 return;
             }
             z = Convert.ToUInt32(x.Substring(2, x.Length-2));
         }
         /// <summary>
         /// Creates a new LMonoidMaxTimes object with a parameter representing a 1000000000 times the value the new object gets.
         /// </summary>
         /// <param name="x">The value the new object gets.</param>
        public LMonoidMaxTimes(uint x) {
            z = x;
            if (z > maxvalue) z = maxvalue;
        }

        /// <summary>
        /// Generates a random LMonoidMaxTimes-element
        /// </summary>
        /// <returns>A new monoid element of a randomly assigned value</returns>
        public static LMonoidMaxTimes Random()
        {
            return new LMonoidMaxTimes((uint)random.Next(0, (int)maxvalue));
        }

        /// <summary>
        /// Creates a new LMonoidMaxTimes object and assigns the provided value.
        /// </summary>
        /// <param name="x">The value the new object gets, if x is less than 0, the value 0 is assigned, if it is greater than 1, the value 1 gets assigned.</param>
        public LMonoidMaxTimes(double x)
        {
            z = (uint)(x*maxvalue);
            if (x > 1) z = maxvalue;
            if (x < 0) z = 0;
        }
        /* Define the necessary operators:*/
        public static bool operator >=(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            return c1.z>=c2.z;
        }
        public static bool operator <=(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            return c1.z <= c2.z;
        }
        public static bool operator >(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            return c1.z > c2.z;
        }
        public static bool operator <(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            return c1.z < c2.z;
        }
       
        /// <summary>
        /// The +-operation of the l-monoid, i.e. the supremum of the lattice.
        /// </summary>
        /// <param name="c1">The first summand.</param>
        /// <param name="c2">The second summand.</param>
        /// <returns>The least element that is bigger or equal to c1 and bigger or equal to c2.</returns>
        public static LMonoidMaxTimes operator +(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            return new LMonoidMaxTimes(c1>=c2?c1:c2);
        }
        /// <summary>
        /// The *-operation of the l-monoid. Here: Multiplication of in R.
        /// </summary>
        /// <param name="c1">The first factor.</param>
        /// <param name="c2">The second factor.</param>
        /// <returns>The least element that is bigger or equal to c1 and bigger or equal to c2.</returns>
        public static LMonoidMaxTimes operator *(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            ulong z = (ulong)c1.z * (ulong)c2.z;
            return new LMonoidMaxTimes((uint)(z/One().z));
        }
        /// <summary>
        /// The infimum-operation of the lattice.
        /// </summary>
        /// <param name="c1">The first element of the infimum.</param>
        /// <param name="c2">The second element of the infimum.</param>
        /// <returns>The greatest element that is smaller or equal to c1 and smaller or equal to c2.</returns>
        public static LMonoidMaxTimes infimum(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            return new LMonoidMaxTimes(c1>=c2?c2:c1);
        }
        /// <summary>
        /// Computes the pseudocomplement c1/c2 denoted c2->c1
        /// </summary>
        /// <param name="c1">The dividend.</param>
        /// <param name="c2">The divisor.</param>
        /// <returns>c2->c1</returns>
        public static LMonoidMaxTimes operator /(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            if (c2 <= c1) return One();
            ulong x = ((ulong)One().z*(ulong)c1.z) / (ulong)c2.z;
            return new LMonoidMaxTimes((uint)x);
        }

        public static bool operator !=(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            return c1.z!=c2.z;
        }
        public static bool operator ==(LMonoidMaxTimes c1, LMonoidMaxTimes c2)
        {
            return c1.z==c2.z;
        }

                
        /// <summary>
        /// Defines how l-monoid-elements should be printed, as well as the one- and the zero-element of the l-monoid.
        /// </summary>
        /// <returns>A string representation of the element.</returns>
        public override string ToString()
        {
            if (this == One()) return "1";
            if (z == 0) return "0";
            int i = (One().z.ToString().Length - z.ToString().Length-1);
            String result = "0.";
            for (int j = 0; j < i; ++j) result += "0";
            result += z.ToString();
            return result;
        }
        
        /// <summary>
        /// Yields the one element of the l-monoid, i.e. the 1 element of *.
        /// </summary>
        /// <returns>A new LMonoidMaxTimes that carries the value 1.</returns>
        public static LMonoidMaxTimes One()
        {
            return new LMonoidMaxTimes(maxvalue);
        }
        
        /// <summary>
        /// Yields the zero element of the l-monoid, i.e. the bottom element of the lattice.
        /// </summary>
        /// <returns>A new LMonoidMaxTimes that carries the value 0.</returns>
        public static LMonoidMaxTimes Zero()
        {
            return new LMonoidMaxTimes(0);
        }

        /// <summary>
        /// Check whether a linear combination exists, using the the algorithm from Zhou / Li: Largest Solution of Linear Equation over the complete heyting algebra.
        /// </summary>
        /// <param name="set">The set of rows the vector shall be combined from.</param>
        /// <param name="vec">The vector (represented as an array) that may be combined linearly.</param>
        /// <returns>A LinearCombination object that indicates in the field CombinationExists, whether a linear combination was found, and, provided a combination was found, in its field Multiplicands the multiplicands that form the linear combination.</returns>
        public static LinearCombination<LMonoidMaxTimes> LinearCombination(List<LMonoidMaxTimes[]> set, LMonoidMaxTimes[] vec)
        {
            LinearCombination<LMonoidMaxTimes> lc;
            lc.CombinationExists = true;
            lc.Multiplicands = new LMonoidMaxTimes[set.Count()];
            for (int j = 0; j < lc.Multiplicands.Length; ++j)
            {
                lc.Multiplicands[j] = One();
                for (int i = 0; i < vec.Length; ++i)
                {
                    //i\in M_j(A,b)?
                    if (!(set[j][i] <= vec[i]))
                        lc.Multiplicands[j] = infimum(lc.Multiplicands[j], vec[i] / set[j][i]);
                }
            }
            //Check, if linear combination was found:
            for (int i = 0; i < vec.Length; ++i)
            {
                if (vec[i] != Matrix<LMonoidMaxTimes>.multiplyVectors(Matrix<LMonoidMaxTimes>.getColumn(set, i), lc.Multiplicands)) lc.CombinationExists = false;
            }
            return lc;
        }


        /// <summary>
        /// Naive check whether a linear combination exists, using the fact that only finitely many vectors can be generated. However, many possible linear combinations are ignored this way, so obviously, optimization is recommended.
        /// </summary>
        /// <param name="set">The set of rows the vector shall be combined from.</param>
        /// <param name="vec">The vector (represented as an array) that may be combined linearly.</param>
        /// <returns>A LinearCombination object that indicates in the field CombinationExists, whether a linear combination was found, and, provided a combination was found, in its field Multiplicands the multiplicands that form the linear combination.</returns>
        public static LinearCombination<LMonoidMaxTimes> LinearCombinationOld(List<LMonoidMaxTimes[]> set, LMonoidMaxTimes[] vec)
        {
            LinearCombination<LMonoidMaxTimes> lc;
            lc.CombinationExists = false;
            lc.Multiplicands = new LMonoidMaxTimes[set.Count()];
            for (int i = 0; i < set.Count(); ++i)
            {
                lc.Multiplicands[i] = Zero();
                if (!lc.CombinationExists) 
                    if (equals(set[i], vec))
                    {
                        lc.Multiplicands[i] = One();
                        lc.CombinationExists=true;
                    }
                }
            return lc;
        }

        /// <summary>
        /// Checks whether two arrays of type LMonoidMaxTimes are equal
        /// </summary>
        /// <param name="a">The first array to be compared.</param>
        /// <param name="b">The second array to be compared.</param>
        /// <returns>true if both arrays are the same length and in each position the have the same l-monoid element.</returns>
        private static bool equals(LMonoidMaxTimes[] a, LMonoidMaxTimes[] b)
        {
            if(b.Length!=a.Length) return false;
            for (int i = 0; i < a.Length; ++i) if (a[i] != b[i]) return false;
            return true;
        }

        public static bool allowedStrings(String s)
        {
            Console.WriteLine(s);
            try
            {
                Convert.ToDouble(s);
                return true;
            }
            catch (Exception)
            {

                return false;
            }
        }
        public static string inputSchema()
        {
            return "uint emptyspace next uint...\nPress Enter for next row";
        }
        

    }
}
