﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Analysis_Part
{
    /// <summary>
    /// The Lattice of all integers, -infinity and +infinity, ordered by the natural order.
    /// </summary>
    [Serializable]
    public class LatticeZ
    {
        /// <summary>
        /// The value of the LatticeZ-object.
        /// </summary>
        private int z;


        public LatticeZ()
        {
    
        }


        /// <summary>
        /// A copy constructor, creating a new LatticeZ object with the same value as the provided one
        /// </summary>
        /// <param name="x">The LatticeZ object to copy.</param>
        public LatticeZ(LatticeZ x)
        {
            z=x.z;
        }

        /// <summary>
        /// Generates a lattice element from a given string. The "inverse" to ToString.
        /// </summary>
        /// <param name="x">The string representing based on which the lattice element may be initialized.</param>
        public LatticeZ(String x)
        {
            if (x == "+" || x == "-")
            {
                if (x == "+") z = Int32.MaxValue;
                else z = Int32.MinValue;
            }
            else
            {
                z = Convert.ToInt32(x);
            }
        }

        /// <summary>
        /// Creates a new LatticeZ object.
        /// </summary>
        /// <param name="x">The value the new object gets.</param>
        public LatticeZ(int x) {
            z = x;
        }

        /* Define the necessary operators:*/
        public static bool operator >=(LatticeZ c1, LatticeZ c2)
        {
            return c1.z>=c2.z;
        }
        public static bool operator <=(LatticeZ c1, LatticeZ c2)
        {
            return c1.z <= c2.z;
        }
        public static bool operator >(LatticeZ c1, LatticeZ c2)
        {
            return c1.z > c2.z;
        }
        public static bool operator <(LatticeZ c1, LatticeZ c2)
        {
            return c1.z < c2.z;
        }
       
        /// <summary>
        /// The +-operation of the lattice, i.e. the supremum.
        /// </summary>
        /// <param name="c1">The first summand.</param>
        /// <param name="c2">The second summand.</param>
        /// <returns>The least element that is bigger or equal to c1 and bigger or equal to c2.</returns>
        public static LatticeZ operator +(LatticeZ c1, LatticeZ c2)
        {
            return new LatticeZ(c1>=c2?c1:c2);
        }
        /// <summary>
        /// The *-operation of the lattice, i.e. the infimum.
        /// </summary>
        /// <param name="c1">The first factor.</param>
        /// <param name="c2">The second factor.</param>
        /// <returns>The greatest element that is smaller or equal to c1 and smaller or equal to c2.</returns>
        public static LatticeZ operator *(LatticeZ c1, LatticeZ c2)
        {
            return new LatticeZ(c1>=c2?c2:c1);
        }
        /// <summary>
        /// Computes the pseudocomplement c1/c2 denoted c2->c1
        /// </summary>
        /// <param name="c1">The dividend.</param>
        /// <param name="c2">The divisor.</param>
        /// <returns>c2->c1</returns>
        public static LatticeZ operator /(LatticeZ c1, LatticeZ c2)
        {
            LatticeZ y = c1;
            LatticeZ x = c2;
            if (x <= y) return One();
            return y;
        }

        public static bool operator !=(LatticeZ c1, LatticeZ c2)
        {
            return c1.z!=c2.z;
        }
        public static bool operator ==(LatticeZ c1, LatticeZ c2)
        {
            return c1.z==c2.z;
        }


        /* Define how Field-elements should be printed, as well as the one- and the zero-element of the Field:*/
        public override string ToString()
        {
            if (z == Int32.MinValue) return "-" + '\u221E';
            if (z == Int32.MaxValue) return "+" + '\u221E';
            return z.ToString();
        }
        /// <summary>
        /// Yields the one element of the lattice, i.e. the top element.
        /// </summary>
        /// <returns>The top-element Int32.MaxValue.</returns>
        public static LatticeZ One()
        {
            return new LatticeZ(Int32.MaxValue);
        }
        /// <summary>
        /// Yields the zero element of the lattice, i.e. the bottom element.
        /// </summary>
        /// <returns>The bottom-element Int32.MinValue.</returns>
        public static LatticeZ Zero()
        {
            return new LatticeZ(Int32.MinValue);
        }

        /// <summary>
        /// Check whether a linear combination exists, using the the algorithm from Zhou / Li: Largest Solution of Linear Equation over the complete heyting algebra.
        /// </summary>
        /// <param name="set">The set of rows the vector shall be combined from.</param>
        /// <param name="vec">The vector (represented as an array) that may be combined linearly.</param>
        /// <returns>A LinearCombination object that indicates in the field CombinationExists, whether a linear combination was found, and, provided a combination was found, in its field Multiplicands the multiplicands that form the linear combination.</returns>
        
        public static LinearCombination<LatticeZ> LinearCombination(List<LatticeZ[]> set, LatticeZ[] vec)
        {

            LinearCombination<LatticeZ> lc;
            lc.CombinationExists = true;
            lc.Multiplicands = new LatticeZ[set.Count()];
            for (int j = 0; j < lc.Multiplicands.Length; ++j)
            {
                lc.Multiplicands[j] = One();
                for (int i = 0; i < vec.Length; ++i)
                {
                    //i\in M_j(A,b)?
                    if (!(set[j][i] <= vec[i]))
                        lc.Multiplicands[j] *= vec[i] / set[j][i];
                }
            }
            //Check, if linear combination was found:
            for (int i = 0; i < vec.Length; ++i)
            {
                if (vec[i] != Matrix<LatticeZ>.multiplyVectors(Matrix<LatticeZ>.getColumn(set, i), lc.Multiplicands)) lc.CombinationExists = false;
            }
            return lc;
        }
        
        

        /// <summary>
        /// Naive check whether a linear combination exists, using the fact that only finitely many vectors can be generated. However, many possible linear combinations are ignored this way, so obviously, optimization is recommended.
        /// </summary>
        /// <param name="set">The set of rows the vector shall be combined from.</param>
        /// <param name="vec">The vector (represented as an array) that may be combined linearly.</param>
        /// <returns>A LinearCombination object that indicates in the field CombinationExists, whether a linear combination was found, and, provided a combination was found, in its field Multiplicands the multiplicands that form the linear combination.</returns>
        public static LinearCombination<LatticeZ> LinearCombinationOld(List<LatticeZ[]> set, LatticeZ[] vec)
        {
            LinearCombination<LatticeZ> lc;
            lc.CombinationExists = false;
            lc.Multiplicands = new LatticeZ[set.Count()];
            for (int i = 0; i < set.Count(); ++i)
            {
                lc.Multiplicands[i] = Zero();
                if (!lc.CombinationExists) 
                    if (equals(set[i], vec))
                    {
                        lc.Multiplicands[i] = One();
                        lc.CombinationExists=true;
                    }
                }
            return lc;
        }

        /// <summary>
        /// Checks whether two arrays of type LatticeZ are equal
        /// </summary>
        /// <param name="a">The first array to be compared.</param>
        /// <param name="b">The second array to be compared.</param>
        /// <returns>true if both arrays are the same length and in each position the have the same lattice element.</returns>
        private static bool equals(LatticeZ[] a, LatticeZ[] b)
        {
            if(b.Length!=a.Length) return false;
            for (int i = 0; i < a.Length; ++i) if (a[i] != b[i]) return false;
            return true;
        }

        public static bool allowedStrings(String s)
        {
            if (s == "-" || s == "+") return true;
            try
            {
                Convert.ToInt32(s);
                return true;
            }
            catch (Exception )
            {
                return false;
            }
        }

        public static string inputSchema()
        {
            return "int next next...\nPress Enter for next row";                 
        }
        
 
        

    }
}
