﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Windows.Forms;
using System.Reflection;
using System.Diagnostics;


namespace Analysis_Part
{
    /// <summary>
    /// A Matrix class that can build matrices from arbitrary types that have an addition and a multiplication operation, as well as a Zero() and a One() function - i.e. Semirings. For some standard types in C#, Zero() and One() are not needed
    /// </summary>
    /// <typeparam name="T">The type over which the matrixes are defined. Operators + and * need to be provided!</typeparam>
    [Serializable]
    public class Matrix<T> where T : class// T needs operators + and *
    {
        public string MatrixName { get; set; }

        private string[] alphabet = new string[0];

        private static int lincombcounter;
        /// <summary>
        /// The matrix-entries.
        /// </summary>
        T[,] values;
        /// <summary>
        /// Creates a new Matrix of a given dimension. The matrix is initialised as a unit matrix with additional zero-rows or zero-columns if it is not symmetric.
        /// </summary>
        /// <param name="rows">The number of rows.</param>
        /// <param name="columns">The number of columns</param>
        public Matrix(int rows, int columns)
        {
            values = new T[rows, columns];
            for (int i = 0; i < rows; ++i) for (int j = 0; j < columns; ++j) values[i, j] = (i == j ? One() : Zero());
        }
        /// <summary>
        /// Creates a new matrix that gets the provided values as matrix-entries.
        /// </summary>
        /// <param name="values">The entries of the matrix.</param>
        public Matrix(T[,] values)
        {
            this.values = values;
        }
        public Matrix(T[] values)
        {
            this.values = new T[values.Length, 1];
            for (int i = 0; i < values.Length; ++i) this.values[i, 0] = values[i];
        }
        public Matrix(Vectoroflmelements<T> values)
        {
            this.values = new T[values.Count(), 1];
            for (int i = 0; i < values.Count(); ++i) this.values[i, 0] = values.get(i);
        }

        public Matrix(List<T[]> values)
        {
            this.values = new T[values.Count, values[0].Length];
            for (int i = 0; i < values.Count; ++i) for (int j = 0; j < values[0].Length; ++j) this.values[i, j] = values[i][j];
        }
        /// <summary>
        /// A copy constructor, copies a given matrix into a new one.
        /// </summary>
        /// <param name="m">The matrix to be copied.</param>
        public Matrix(Matrix<T> m)
        {
            values = new T[m.rows(), m.columns()];
            for (int i = 0; i < rows(); ++i) for (int j = 0; j < columns(); ++j) values[i, j] = m.get(i, j);
        }
        /// <summary>
        /// Set an individual entry of the matrix to a provided value.
        /// </summary>
        /// <param name="row">The row index of the entry to be set.</param>
        /// <param name="column">The column index of the entry to be set.</param>
        /// <param name="value">The value the entry gets.</param>
        public void set(int row, int column, T value)
        {
            values[row, column] = value;
        }
        /// <summary>
        /// Get an individual entry of the matrix.
        /// </summary>
        /// <param name="row">The row index of the entry to be read.</param>
        /// <param name="column">The column index of the entry to be read.</param>
        /// <returns>The value at index (i,j).</returns>
        public T get(int row, int column)
        {
            return values[row, column];
        }
        /// <summary>
        /// Gives the number of rows the matrix has.
        /// </summary>
        /// <returns>The number of rows the matrix has.</returns>
        public int rows()
        {
            return values.GetUpperBound(0) + 1;
        }
        /// <summary>
        /// Gives the number of columns the matrix has.
        /// </summary>
        /// <returns>The number of columns the matrix has.</returns>
        public int columns()
        {
            return values.GetUpperBound(1) + 1;
        }

        /// <summary>
        /// set & get for the alphabet of the automata
        /// </summary>
        /// <param name="alpha"></param>
        public void setAlphabet(string [] alpha)
        {
            this.alphabet =  new string [alpha.Length];
            for (int i = 0; i < alpha.Length; ++i) alphabet[i] = alpha[i];
        }

        public string[] getAlphabet()
        {
            return this.alphabet;
        }

        /// <summary>
        /// Multiplies two vectors, given as arrays.
        /// </summary>
        /// <param name="x">The column vector to be multiplied.</param>
        /// <param name="y">The row vector to be multiplied.</param>
        /// <returns>The product x*y.</returns>
        public static T multiplyVectors(T[] x, T[] y)
        {
            T result = (dynamic)x[0] * (dynamic)y[0];
            int upper = x.GetUpperBound(0);
            if (upper > y.GetUpperBound(0)) upper = y.GetUpperBound(0);
            for (int i = 1; i <= upper; ++i) result += (dynamic)x[i] * (dynamic)y[i];
            return result;
        }


        /// <summary>
        /// \times-multiplies two vectors, given as arrays.
        /// </summary>
        /// <param name="x">The column vector to be multiplied.</param>
        /// <param name="y">The row vector to be multiplied.</param>
        /// <returns>The product x\otimes y.</returns>
        public static T divideVectors(T[] x, T[] y)
        {
            T result = (dynamic)y[0] / (dynamic)x[0];
            for (int i = 1; i <= x.GetUpperBound(0); ++i) result *= ((dynamic)y[i] / (dynamic)x[i]);
            return result;
        }

        /// <summary>
        /// Yields the i-th row in form of an array.
        /// </summary>
        /// <param name="i">The index of the row to be read.</param>
        /// <returns>The i-th row of the matrix.</returns>
        public T[] row(int i)
        {
            T[] result = new T[columns()];
            for (int j = 0; j < columns(); ++j) result[j] = get(i, j);
            return result;
        }
        /// <summary>
        /// Yields the i-th column in form of an array.
        /// </summary>
        /// <param name="i">The index of the column to be read.</param>
        /// <returns>The i-th column of the matrix.</returns>
        public T[] column(int i)
        {
            T[] result = new T[rows()];
            for (int j = 0; j < rows(); ++j) result[j] = get(j, i);
            return result;
        }
        /// <summary>
        /// Adds two complex matrices.
        /// </summary>
        /// <param name="m1">The first summand.</param>
        /// <param name="m2">The second summand.</param>
        /// <returns>The sum of m1 and m2.</returns>
        public static Matrix<T> operator +(Matrix<T> m1, Matrix<T> m2)
        {
            T[,] values = new T[m1.rows(), m1.columns()];
            for (int i = 0; i < m1.rows(); ++i)
            {
                for (int j = 0; j < m1.columns(); ++j)
                {
                    values[i, j] = (dynamic)m1.get(i, j) + (dynamic)m2.get(i, j);
                }
            }
            return new Matrix<T>(values);
        }
        /// <summary>
        /// Multiplies two matrices. Note that the row count of the first matrix needs to equal the column count of the second matrix.
        /// </summary>
        /// <param name="m1">The first factor.</param>
        /// <param name="m2">The second factor.</param>
        /// <returns>The product of m1 and m2.</returns>
        public static Matrix<T> operator *(Matrix<T> m1, Matrix<T> m2)
        {
            T[,] values = new T[m1.rows(), m2.columns()];
            for (int i = 0; i < m1.rows(); ++i)
            {
                for (int j = 0; j < m2.columns(); ++j)
                {
                    values[i, j] = multiplyVectors(m1.row(i), m2.column(j));
                }
            }
            return new Matrix<T>(values);
        }

        /// <summary>
        /// \otimes-multiplies two matrices. Note that the row count of the first matrix needs to equal the column count of the second matrix.
        /// </summary>
        /// <param name="m1">The first factor.</param>
        /// <param name="m2">The second factor.</param>
        /// <returns>The product of m1 and m2.</returns>
        public static Matrix<T> operator /(Matrix<T> m1, Matrix<T> m2)
        {
            T[,] values = new T[m1.rows(), m2.columns()];
            for (int i = 0; i < m1.rows(); ++i)
            {
                for (int j = 0; j < m2.columns(); ++j)
                {
                    values[i, j] = divideVectors(m1.row(i), m2.column(j));
                }
            }
            return new Matrix<T>(values);
        }

        public static Matrix<T> transpose(Matrix<T> m)
        {
            Matrix<T> result = new Matrix<T>(m.columns(), m.rows());
            for (int i = 0; i < m.rows(); ++i)
                for (int j = 0; j < m.columns(); ++j)
                    result.set(j, i, m.get(i, j));
            return result;
        }

        public static Matrix<T> elementwiseMult(Matrix<T> m1, Matrix<T> m2)
        {
            Matrix<T> result = new Matrix<T>(m1.rows(), m1.columns());
            for (int i = 0; i < m1.rows(); ++i)
                for (int j = 0; j < m1.columns(); ++j)
                    result.set(i, j, (dynamic)m1.get(i, j) * m2.get(i, j));
            return result;
        }

        /// <summary>
        /// Compares two matrices element wise and yields true if and only if each entry is equal.
        /// </summary>
        /// <param name="m1">The first matrix to be compared.</param>
        /// <param name="m2">The second matrix to be compared.</param>
        /// <returns>True if the matrices are equal, false otherwise.</returns>
        public static bool operator ==(Matrix<T> m1, Matrix<T> m2)
        {
            if (m1.rows() != m2.rows() || m1.columns() != m2.columns()) return false;
            for (int i = 0; i < m1.rows(); ++i)
            {
                for (int j = 0; j < m2.columns(); ++j)
                {
                    if ((dynamic)m1.get(i, j) != (dynamic)m2.get(i, j)) return false;
                }
            }
            return true;
        }


        public static bool operator <=(Matrix<T> m1, Matrix<T> m2)
        {
            if (m1.rows() != m2.rows() || m1.columns() != m2.columns()) return false;


            for (int i = 0; i < m1.rows(); ++i)
            {
                LinearCombination<T> lc;
                T[] vec = m1.row(i);

                try
                {
                    Type t = typeof(T);
                    lc = (dynamic)t.GetMethod("LinearCombination").Invoke(null, new object[] { m2.toList(), vec });
                }
                catch (Exception)
                {
                    //Console.WriteLine("LinearCombination-Methode existiert nicht");
                    lc = new LinearCombination<T>();
                    lc.CombinationExists = false;
                }
                if (!lc.CombinationExists) return false;
            }
            return true;
        }

        public static bool operator >=(Matrix<T> m1, Matrix<T> m2)
        {
            if (m1.rows() != m2.rows() || m1.columns() != m2.columns()) return false;


            for (int i = 0; i < m1.rows(); ++i)
            {
                LinearCombination<T> lc;
                T[] vec = m2.row(i);

                try
                {
                    Type t = typeof(T);
                    lc = (dynamic)t.GetMethod("LinearCombination").Invoke(null, new object[] { m1.toList(), vec });
                }
                catch (MissingMethodException)
                {
                    //Console.WriteLine("LinearCombination-Methode existiert nicht");
                    //lc = AskUserForLinearCombination(m1.toList(), vec);
                    lc = new LinearCombination<T>();
                    lc.CombinationExists = false;
                    // abort computing step:

                }
                catch (Exception e)
                {
                    string caption = "An exception occurs while method-Invoke!";
                    string message = e.ToString();
                    MessageBoxButtons buttons = MessageBoxButtons.OK;
                    DialogResult result;

                    // Displays the exception in the dialog box
                    result = MessageBox.Show(message, caption, buttons);

                    lc = new LinearCombination<T>();
                    lc.CombinationExists = false;
                    // abort computing step:
                }
                if (!lc.CombinationExists) return false;
            }
            return true;
        }
        /// <summary>
        /// Compares two matrices element wise and yields true if and only if each entry is not equal.
        /// </summary>
        /// <param name="m1">The first matrix to be compared.</param>
        /// <param name="m2">The second matrix to be compared.</param>
        /// <returns>False if the matrices are equal, true otherwise.</returns>
        public static bool operator !=(Matrix<T> m1, Matrix<T> m2)
        {
            return !(m1 == m2);
        }

        /// <summary>
        /// Converts a matrix into a string.
        /// </summary>
        /// <returns>A string representation of the matrix.</returns>
        public override string ToString()
        {
            string s = "";
            for (int i = 0; i < rows(); ++i)
            {
                for (int j = 0; j < columns(); ++j)
                {
                    s = s + get(i, j) + "  ";
                }
                s += '\n';
            }
            return s;
        }

        /// <summary>
        /// Yields the one of the semiring. Either T needs to be of type Semiring or int, uint, double for this to work.
        /// </summary>
        /// <returns>The one of the given semiring.</returns>
        public static T One()
        {
            if ((typeof(T) == typeof(int)) || (typeof(T) == typeof(uint)) || (typeof(T) == typeof(double))) return (dynamic)1;
            try
            {
                Type t = typeof(T);

                return (dynamic)t.GetMethod("One").Invoke(null, null);
            }
            catch (Exception e)
            {
                throw (new Exception("One for type " + typeof(T).ToString() + " not defined! Define function One() in " + typeof(T).ToString() + " that yields the one of said semiring."));
            };


        }
        /// <summary>
        /// Yields the zero of the semiring. Either T needs to be of type Semiring or int, uint, double for this to work.
        /// </summary>
        /// <returns>The zero of the given semiring.</returns>

        public static T Zero()
        {
            if ((typeof(T) == typeof(int)) || (typeof(T) == typeof(uint)) || (typeof(T) == typeof(double))) return (dynamic)0;
            try
            {
                Type t = typeof(T);

                return (dynamic)t.GetMethod("Zero").Invoke(null, null);
            }
            catch (Exception e)
            {
                throw (new Exception("Zero for type " + typeof(T).ToString() + " not defined! Define function Zero() in " + typeof(T).ToString() + " that yields the zero of said semiring."));
            };


        }
        /// <summary>
        /// Yields a list of T arrays representing the rows of the matrix.
        /// </summary>
        /// <returns>A list of T arrays representing the rows of the matrix.</returns>
        public List<T[]> toList()
        {
            List<T[]> result = new List<T[]>();
            for (int i = 0; i < rows(); ++i)
            {
                result.Add(row(i));
            }
            return result;
        }
        /// <summary>
        /// Yields the i-th unit vector of a given dimension in form of a T-array.
        /// </summary>
        /// <param name="index">The index where the one should be.</param>
        /// <param name="dimension">The dimension of the vector</param>
        /// <returns>The index-th unit vector of dimension 'dimension' in form of a T-array.</returns>
        public static T[] UnitVector(int index, int dimension) // Yields 'index'-th unit-vector of dimension 'dimension'
        {
            T[] result = new T[dimension];
            for (int i = 0; i < dimension; ++i)
                if (i == index - 1) result[i] = One(); else result[i] = Zero();
            return result;
        }
        /// <summary>
        /// Unifinished! Ask the user for a linear combination in case no linear combination function is provided by the type T.
        /// If a method for linear combination exists, the user should also implement that for every semiring. Calculation by hand
        /// would be to costly!
        /// </summary>
        /// <param name="set">The set of row vectors, represented as a T-array from which the combination may be formed.</param>
        /// <param name="vec">The vector to be formed via linear combination.</param>
        /// <returns>A linear combination object that carries the provided multiplicands if the user answered a linear combination does exist. The boolean CombinationExists is set accordingly.</returns>
        public static LinearCombination<T> AskUserForLinearCombination(List<T[]> set, T[] vec) // Ask user if vec is linear combination of set and if it is, what the linear combination is
        {//TODO!! Just a stub!
            LinearCombination<T> lc = new LinearCombination<T>();
     
            /* Console.WriteLine("Automatic check if a vector is a linear combination of another is not implemented yet. Therefore the user must provide the necessary linear combinations. Does a linear combination of these vectors: ");
             //ausgeben von set
             Console.WriteLine("exist, such that it yields");
             //ausgeben von vec
             Console.WriteLine("Input 'y' if it exists, 'n' otherwise.");
             string answer = Console.ReadLine();
             while (answer != "y" && answer != "n") answer = Console.ReadLine();
             if (answer == "n") { lc.CombinationExists = false; return lc; }
             lc.Multiplicands = new T[set.Count()];
             lc.CombinationExists = true;
             for (int i = 0; i < set.Count(); ++i)
             {
                 Console.WriteLine("What is the multiplicand for ");
                 Console.WriteLine(set[i]);
                 answer = Console.ReadLine();
                 // Speichern von lc.Multiplicands[i]=answer
             }
             */
            //GUI
            bool inputRequirement = true;
            string setout = "";
            for (int i = 0; i < set.Count(); ++i)
            {
                for (int j = 0; j < set[i].Count(); ++j)
                {
                    setout += set[i][j].ToString()+" ";
                }
                setout += "\n";
            }
            setout += "Vector to combine= \n";
            for (int j = 0; j < vec.Count(); ++j)
            {
                setout += vec[j].ToString() + " ";
            }

             
            Microsoft.VisualBasic.MsgBoxResult response = Microsoft.VisualBasic.MsgBoxResult.Yes;
             //response = Microsoft.VisualBasic.Interaction.MsgBox("Automatic check if a vector is a linear combination of another is not implemented yet. Therefore the user must provide the necessary linear combinations. \n Does a linear combination of these vectors:"+
             //"\n"+setout+ "\n"+ "exist?", Microsoft.VisualBasic.MsgBoxStyle.YesNo, "Linear Combination?");
            if (response != Microsoft.VisualBasic.MsgBoxResult.Yes)
            {
                inputRequirement = false;
            }
             if (inputRequirement)
             {
                 lc.Multiplicands = new T[set.Count()];
                 lc.CombinationExists = true;
                 string s = (Microsoft.VisualBasic.Interaction.InputBox("Please type in the solution vector of size "+set.Count +":", "LinearCombination"));
                 if (String.IsNullOrWhiteSpace(s)) { lc.CombinationExists = false; }
                 // TODO only naive solution to enable an input, but without usibility adaption for errors in the input 
                 else
                 {
                     string[] vector = s.Split(' ');
                     for (int j = 0; j < set.Count(); ++j)
                     {
                         try
                         {
                             T oItem = Activator.CreateInstance(typeof(T), vector[j]) as T;
                             lc.Multiplicands[j] = oItem;
                         }
                         catch { 
                             MessageBox.Show("Your input is wrong!");
                             lc.CombinationExists = false;
                         }
                     }
                 }
             }

            return lc;
        }

        public static T[] getColumn(List<T[]> set, int i)
        {
            T[] result = new T[set.Count()];
            for (int j = 0; j < set.Count(); ++j) result[j] = set[j][i];
            return result;
        }
        /// <summary>
        /// Yields a copy of a given list of arrays with the exception of index i.
        /// </summary>
        /// <typeparam name="T">The type of the arrays in the list.</typeparam>
        /// <param name="list">The list to be copied.</param>
        /// <param name="i">The index of the list-entry to be left out</param>
        /// <returns>A copy of list without the entry at position i.</returns>
        public static List<T[]> ListExcept<T>(List<T[]> list, int i)
        {
            List<T[]> result = new List<T[]>();
            for (int j = 0; j < i; ++j) result.Add(list[j]);
            for (int j = i + 1; j < list.Count(); ++j) result.Add(list[j]);
            return result;
        }


        /// <summary>
        /// Yields a matrix that is a copy of a provided matrix sans the i-th row.
        /// </summary>
        /// <param name="m">The matrix to be copied.</param>
        /// <param name="i">The row to be left out.</param>
        /// <returns>A copy of m sans row i.</returns>
        public static Matrix<T> MatrixExcept(Matrix<T> m, int i)
        {
            Matrix<T> result = new Matrix<T>(m.rows() - 1, m.columns());
            for (int j = 0; j < i; ++j) for (int k = 0; k < m.columns(); ++k) result.set(j, k, m.get(j, k));
            for (int j = i + 1; j < m.rows(); ++j) for (int k = 0; k < m.columns(); ++k) result.set(j - 1, k, m.get(j, k));
            return result;
        }

        /// <summary>
        /// Factorizes a given matrix of type T using the LinearCombination-method of said type, provided it exists. Asks the user for the linear combination instead if it does not exist.
        /// </summary>
        /// <param name="m">The matrix to be factorized.</param>
        /// <param name="n">If only a certain number of rows (from the top) shall be checked whether they can be factorized away, this number indicates how many rows to consider. '-1' means to consider all rows.</param>
        /// <returns>An array of two matrices: The epi-part (index 0) and the monic part (index 1) of the factorization.</returns>
        public static Matrix<T>[] factorize(Matrix<T> m, int n = -1)
        // Parameter n: If non-negative, only try to eliminate the first n rows.
        {
            if (n == -1) n = m.rows();
            Matrix<T>[] results = new Matrix<T>[2];
            Matrix<T> epi = new Matrix<T>(m);
            //List<T[]> mono = new List<T[]>();
            Matrix<T> mono = new Matrix<T>(m.rows(), m.rows());

            for (int i = 0; (i < epi.rows() && epi.rows() > 1 && n > i + m.rows() - epi.rows()); ++i)
            //for (int i = 0; (i < epi.rows()) && (epi.rows() > 1); ++i)
            {

                LinearCombination<T> lc;
                T[] vec = epi.row(i);
                Matrix<T> workingCopy = MatrixExcept(epi, i);

                try
                {
                    Type t = typeof(T);
                    lc = (dynamic)t.GetMethod("LinearCombination").Invoke(null, new object[] { workingCopy.toList(), vec });
                    ++lincombcounter;
                    //Console.WriteLine("LinearCombination-Methode existiert!");
                }
                catch (Exception)
                {
                    //Console.WriteLine("LinearCombination-Methode existiert nicht");
                    lc = lc = new LinearCombination<T>();
                    lc.CombinationExists = false;
                }

                if (lc.CombinationExists)
                {
                    epi = workingCopy;
                    Matrix<T> mono2 = new Matrix<T>(mono.columns(), mono.columns() - 1);
                    for (int j = 0; j < i; ++j)
                        for (int k = 0; k < mono2.columns(); ++k)
                            mono2.set(j, k, (j == k ? One() : Zero()));
                    for (int k = 0; k < mono2.columns(); ++k) mono2.set(i, k, lc.Multiplicands[k]);
                    for (int j = i + 1; j < mono2.rows(); ++j)
                        for (int k = 0; k < mono2.columns(); ++k)
                            mono2.set(j, k, (j == k + 1 ? One() : Zero()));


                    mono *= mono2;
                    --i;
                }
            }
            results[0] = epi;
            results[1] = mono;
            return results;
        }



        /*public static List<T[]>[] factorize(Matrix<T> m)
        {
            List<T[]>[] results = new List<T[]>[2];
            List<T[]> epi = m.toList();
            //List<T[]> mono = new List<T[]>();
            Matrix<T> mono = new Matrix<T>(m.rows(), m.columns());
            for (int i = 0; (i < epi.Count()&& epi.Count()>1); ++i)
            {
                
                LinearCombination<T> lc;
                Console.WriteLine("Check!" + i);
                T[] vec = epi.ElementAt<T[]>(i);
                List<T[]> workingCopy = ListExcept<T>(epi, i);
                    
                try
                {
                    Type t = typeof(T);
                    lc = (dynamic)t.GetMethod("LinearCombination").Invoke(null, new object[] { workingCopy, vec });
                    //Console.WriteLine("LinearCombination-Methode existiert!");
                }
                catch (Exception)
                {
                    //Console.WriteLine("LinearCombination-Methode existiert nicht");
                    lc=AskUserForLinearCombination(workingCopy, vec);
                }

                if (lc.CombinationExists)
                {
                    epi = workingCopy;
                    Matrix<T> mono2 = new Matrix<T>(mono.rows(), mono.rows() - 1);
                    for (int j = 0; j < i; ++j)
                        for (int k = 0; k < mono2.columns(); ++k)
                            mono2.set(j, k, (j == k ? One() : Zero()));
                    for (int k = 0; k < mono2.columns(); ++k) mono2.set(i, k, lc.Multiplicands[k]);
                    for (int j = i + 1; j < mono2.rows(); ++j)
                        for (int k = 0; k < mono2.columns(); ++k)
                            mono2.set(j, k, (j == k+1 ? One() : Zero()));
                    mono *= mono2;
                    --i;
                }
            }
            results[0] = epi;
            results[1] = mono.toList();
            return results;
        }
        public List<T[]>[] factorize()
        {
            return(factorize(this));
        }*/
        /// <summary>
        /// Factorizes the matrix using the LinearCombination-method of T, provided it exists. Asks the user for the linear combination instead if it does not exist.
        /// </summary>
        /// <param name="n">If only a certain number of rows (from the top) shall be checked whether they can be factorized away, this number indicates how many rows to consider. '-1' means to consider all rows.</param>
        /// <returns>An array of two matrices: The epi-part (index 0) and the monic part (index 1) of the factorization.</returns>
        public Matrix<T>[] factorize(int n = -1)
        {
            return (factorize(this, n));
        }

        /// <summary>
        /// Applies functor F from paper 'preorder' to a given matrix.
        /// </summary>
        /// <param name="e">The matrix where the functor F shall be applied.</param>
        /// <param name="alphabetsize">The size of the alphabet in use, needed to determine the concrete form of F.</param>
        /// <returns>The matrix that F yields when applied to e.</returns>
        public static Matrix<T> F(Matrix<T> e, int alphabetsize)
        {
            Matrix<T> Fe = new Matrix<T>(e.rows() * alphabetsize + 1, e.columns() * alphabetsize + 1);
            for (int a = 0; a < alphabetsize; ++a)
                for (int x = 0; x < e.columns(); ++x)
                    for (int b = 0; b < alphabetsize; ++b)
                        for (int y = 0; y < e.rows(); ++y)
                            Fe.set(y * alphabetsize + b, x * alphabetsize + a, ((a == b) ? e.get(y, x) : Zero()));
            for (int r = 0; r < Fe.rows() - 1; ++r) Fe.set(r, Fe.columns() - 1, Zero());
            for (int c = 0; c < Fe.columns() - 1; ++c) Fe.set(Fe.rows() - 1, c, Zero());
            Fe.set(Fe.rows() - 1, Fe.columns() - 1, One());
            return Fe;
        }
        /// <summary>
        /// Applies functor F from paper 'preorder' to the matrix.
        /// </summary>
        /// <param name="alphabetsize">The size of the alphabet in use, needed to determine the concrete form of F.</param>
        /// <returns>The matrix that F yields when applied to the matrix.</returns>
        public Matrix<T> F(int alphabetsize)
        {
            return F(this, alphabetsize);
        }
        /// <summary>
        /// Applies Algorithm B from paper 'preorder' to the matrix.
        /// </summary>
        /// <returns>An array that contains two matrixes: The final e matrix and the final m matrix of Algorithm B.</returns>
        public List<Matrix<T>> AlgorithmB(bool complete)
        {
            return AlgorithmB(this, complete);
        }
         
        /// <summary>
        /// Applies Algorithm B from paper 'preorder' to a given matrix.
        /// </summary>
        /// <param name="alpha">The matrix Algorithm B shall be applied to.</param>
        /// <returns>An array that contains two matrixes: The final e matrix and the final m matrix of Algorithm B.</returns>
        public static List<Matrix<T>> AlgorithmB(Matrix<T> alpha, bool complete)
        {
            Console.WriteLine(alpha);
            lincombcounter = 0;
            int X = alpha.columns();
            int A = (alpha.rows() - 1) / alpha.columns();
            Matrix<T> ei;
            Matrix<T> eip1 = new Matrix<T>(0, X);
            List<Matrix<T>> resultList = new List<Matrix<T>>();
            Matrix<T>[] result;
            int i = 1;
            do
            {
                ei = eip1;
                Matrix<T> d = F(ei, A) * alpha;     
                result = factorize(d);
                eip1 = result[0];
                if (complete && (ei != eip1)) resultList.Add(d);
            
                if (complete && (ei != eip1))
                {
                    resultList.Add(result[1]);
                    resultList.Add(result[0]);
                }
                //only last result
                if (ei >= eip1)
                {
                    resultList.Add(d);
                    resultList.Add(result[1]);
                    resultList.Add(result[0]);
                }


            } while (!(ei >= eip1));
            Console.WriteLine("Anzahl der Gaussschritte: " + lincombcounter);
            return resultList;
        }

        /// <summary>
        /// Applies Algorithm B from paper 'preorder' to the matrix, using an optimization applicable to fields and lattices if we only check for equality of rows instead of linear combinations.
        /// </summary>
        /// <returns>An array that contains two matrixes: The final e matrix and the final m matrix of Algorithm B.</returns>
        public Matrix<T>[] OptimizedAlgorithmB()
        {
            return OptimizedAlgorithmB(this);
        }
        /// <summary>
        /// Applies Algorithm B from paper 'preorder' to a given matrix, using an optimization applicable to fields and lattices if we only check for equality of rows instead of linear combinations.
        /// </summary>
        /// <param name="alpha">The matrix Algorithm B shall be applied to.</param>
        /// <returns>An array that contains two matrixes: The final e matrix and the final m matrix of Algorithm B.</returns>
        public static Matrix<T>[] OptimizedAlgorithmB(Matrix<T> alpha)
        // Optimization ONLY applicable in case it suffices to check new rows for redundancy
        {
            int X = alpha.columns();
            int A = (alpha.rows() - 1) / alpha.columns();
            Matrix<T> ei;
            Matrix<T> eip1 = new Matrix<T>(0, X);
            Matrix<T>[] result;
            int i = 1;
            do
            {
                ei = eip1;

                Matrix<T> d = F(ei, A) * alpha;
                Console.WriteLine("d" + i);
                Console.WriteLine(d);
                result = factorize(d, d.rows() - ei.rows());
                eip1 = result[0];
                Console.WriteLine("m" + i);
                Console.WriteLine(result[1]);
                Console.WriteLine("e" + i++);
                Console.WriteLine(result[0]);

            } while (ei != eip1 );
            return result;
        }

        public int alphabetsize()
        {
            return (rows() - 1) / columns();
        }

        public T[] getRow(int i)
        {
            T[] result = new T[columns()];
            for (int j = 0; j < columns(); ++j) result[j] = values[i, j];
            return result;
        }

        public T[] getFinalWeightRow()
        {
            return getRow(rows() - 1);
        }

        public void setRow(int i, T[] row, bool accountForAdditionalEntry = false)
        {
            for (int j = 0; j < columns() - (accountForAdditionalEntry ? 1 : 0); ++j) values[i, j] = row[j];

        }

        public Matrix<T> getTransitionsForSymbol(int i, bool accountForAdditionalEntry = false)
        {
            Matrix<T> result = new Matrix<T>(columns() + (accountForAdditionalEntry ? 1 : 0), columns() + (accountForAdditionalEntry ? 1 : 0));
            for (int j = i; j < rows() - 1; j += alphabetsize())
                result.setRow(j / alphabetsize(), getRow(j), accountForAdditionalEntry);
            if (accountForAdditionalEntry)
            {
                for (int j = 0; j < columns(); ++j)
                {
                    result.set(j, columns(), Zero());
                    result.set(columns(), j, Zero());
                }
                result.set(columns(), columns(), One());
            }
            return result;
        }
        /// <summary>
        /// Assumes, that the matrix is a column vector (if it is not, ist just takes the first column) and transforms this vector into a T-array
        /// </summary>
        /// <returns>A T-array that contains (from top to bottom) all entries of the 0th column.</returns>
        public T[] ColumnVecToList()
        {
            T[] result = new T[rows()];
            for (int i = 0; i < rows(); ++i) result[i] = values[i, 0];
            return result;
        }

        public bool checkInCongruenceClosure(List<T[]> R, T[] x)
        {
            int n = x.Length / 2;
            T[] x1 = new T[n];
            T[] x2 = new T[n];
            for (int i = 0; i < n; ++i)
            {
                x1[i] = x[i];
                x2[i] = x[n + i];
            }
            foreach (T[] y in R)
            {
                bool flag = true;
                for (int i = 0; i < n; ++i)
                    if (!((dynamic)y[i] <= x[i]))
                    {
                        flag = false;
                        break;
                    }
                if (flag)
                    for (int i = 0; i < n; ++i)
                        x2[i] = (dynamic)x2[i] + y[i];
                flag = true;
                for (int i = n; i < x.Length; ++i)
                    if (!((dynamic)y[i] <= x[i]))
                    {
                        flag = false;
                        break;
                    }
                if (flag)
                    for (int i = 0; i < n; ++i)
                        x1[i] = (dynamic)x1[i] + y[i + n];
            }
            for (int i = 0; i < n; ++i)
                if ((dynamic)x1[i] != x2[i]) return false;
            return true;
        }

        public bool bonchiVectors(T[] x, T[] y, bool idempotence = false)
        {
            List<String> wordlist = new List<String>();
            List<String> wordset = new List<String>();
            lincombcounter = 0;
            Type t = typeof(T);
            LinearCombination<T> lc;
            List<T[]> R = new List<T[]>();
            List<T[]> todo = new List<T[]>();
            int n = 2 * columns();
            T[] initial = new T[n];
            T[] initialR = new T[n];
            T[] initialReflexiveX = new T[n];
            T[] initialReflexiveY = new T[n];
            for (int i = 0; i < columns(); ++i) initial[i] = initialR[i + columns()] = initialReflexiveX[i] = initialReflexiveX[i + columns()] = x[i];
            for (int i = 0; i < columns(); ++i) initial[i + columns()] = initialR[i] = initialReflexiveY[i] = initialReflexiveY[i + columns()] = y[i];
            if (!idempotence)
            {
                R.Add(initialReflexiveX);
                lc = (dynamic)t.GetMethod("LinearCombination").Invoke(null, new object[] { R, initialReflexiveY });
                if (lc.CombinationExists) R.Add(initialReflexiveY);
                //++lincombcounter;
            }
            todo.Add(initial);
            todo.Add(initialR);
            for (int i = 0; i < todo.Count; ++i) wordlist.Add("");
            wordset.Add("");
            while (!(todo.Count() == 0))
            {
                String currentword = wordlist.ElementAt(0);
                wordlist.RemoveAt(0);
                //if(R.Count>0) Console.WriteLine("Größe von R: " + R.Count+", Erstes Element in R hat so viele Einträge: "+R[0].Length);
                T[] candidate = todo[0];
                todo.Remove(candidate);
                if (idempotence)
                {
                    //++lincombcounter;
                    if (checkInCongruenceClosure(R, candidate)) continue;
                }
                else
                {
                    lc = (dynamic)t.GetMethod("LinearCombination").Invoke(null, new object[] { R, candidate });
                    //++lincombcounter;
                    if (lc.CombinationExists) continue;
                }
                ++lincombcounter;
                R.Add(candidate);
                if (!wordset.Contains(currentword)) wordset.Add(currentword);
                T[] candidateleft = new T[columns()];
                T[] candidateright = new T[columns()];
                for (int i = 0; i < columns(); ++i)
                {
                    candidateleft[i] = candidate[i];
                    candidateright[i] = candidate[i + columns()];
                }
                T[] final = getFinalWeightRow();
                if ((dynamic)(multiplyVectors(candidateleft, final)) != multiplyVectors(candidateright, final))
                {
                    Console.WriteLine("The following words were explored: ");
                    for (int i = 0; i < wordset.Count; ++i) Console.WriteLine(wordset[i] == "" ? "epsilon" : wordset[i]);
                    return false;
                }
                char letter = 'a';
                for (int i = 0; i < alphabetsize(); ++i)
                {
                    Matrix<T> M = getTransitionsForSymbol(i);
                    T[] candidateleftM = (M * (new Matrix<T>(candidateleft))).ColumnVecToList();
                    T[] candidaterightM = (M * (new Matrix<T>(candidateright))).ColumnVecToList();
                    T[] LR = new T[n];
                    T[] RL = new T[n];
                    T[] LL = new T[n];
                    T[] RR = new T[n];
                    for (int j = 0; j < columns(); ++j) LR[j] = RL[j + columns()] = LL[j] = LL[j + columns()] = candidateleftM[j];
                    for (int j = 0; j < columns(); ++j) RL[j] = LR[j + columns()] = RR[j] = RR[j + columns()] = candidaterightM[j];
                    todo.Add(LR);
                    todo.Add(RL);
                    String wordtoadd = currentword + letter++;
                    wordlist.Add(wordtoadd);
                    wordlist.Add(wordtoadd);
                    //Reflexive pairs may be added to ensure that transitive pairs are not unnecessarily explored, however, they need not be explored any further, so they get directly added to R, instead of todo.
                    //This is only necessary if the semiring is not additeively idempotent.
                    if (!idempotence)
                    {
                        lc = (dynamic)t.GetMethod("LinearCombination").Invoke(null, new object[] { R, LL });
                        if (!lc.CombinationExists) R.Add(LL);
                        //++lincombcounter;
                        lc = (dynamic)t.GetMethod("LinearCombination").Invoke(null, new object[] { R, RR });
                        if (!lc.CombinationExists) R.Add(RR);
                        //++lincombcounter;
                    }

                }
            }
            Console.WriteLine("The following words were explored: ");
            for (int i = 0; i < wordset.Count; ++i) Console.WriteLine(wordset[i] == "" ? "epsilon" : wordset[i]);
            return true;
        }

        /// <summary>
        /// An algorithm based on Bonchi's work using rewriting. Ignore for Master thesis
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public bool bonchiVectorsRewriting(T[] x, T[] y)
        {
            List<String> wordlist = new List<String>();
            List<String> wordset = new List<String>();
            lincombcounter = 0;
            Type t = typeof(T);
            RuleSystem<T> R = new RuleSystem<T>();
            List<Rule<T>> todo = new List<Rule<T>>();
            Rule<T> initial = new Rule<T>();
            initial.Left = new Vectoroflmelements<T>(x);
            initial.Right = new Vectoroflmelements<T>(y);
            todo.Add(initial);
            wordlist.Add("");
            wordset.Add("");
            while (!(todo.Count() == 0))
            {
                String currentword = wordlist.ElementAt(0);
                wordlist.RemoveAt(0);
                //if(R.Count>0) Console.WriteLine("Größe von R: " + R.Count+", Erstes Element in R hat so viele Einträge: "+R[0].Length);
                Rule<T> candidate = todo[0];
                todo.Remove(candidate);
                if (R.ApplyAtOnce(candidate.Left) == R.ApplyAtOnce(candidate.Right)) continue;

                Rule<T> candidateL = new Rule<T>();
                candidateL.Left = candidate.Left;
                candidateL.Right = candidate.Left + candidate.Right;
                R.Add(candidateL);
                Rule<T> candidateR = new Rule<T>();
                candidateR.Left = candidate.Right;
                candidateR.Right = candidate.Left + candidate.Right;
                R.Add(candidateR);


                if (!wordset.Contains(currentword)) wordset.Add(currentword);
                T[] candidateleft = candidate.Left.elements.ToArray<T>();
                T[] candidateright = candidate.Right.elements.ToArray<T>();
                T[] final = getFinalWeightRow();
                if ((dynamic)(multiplyVectors(candidateleft, final)) != multiplyVectors(candidateright, final))
                {
                    Console.WriteLine("The following words were explored: ");
                    for (int i = 0; i < wordset.Count; ++i) Console.WriteLine(wordset[i] == "" ? "epsilon" : wordset[i]);
                    return false;
                }
                char letter = 'a';
                for (int i = 0; i < alphabetsize(); ++i)
                {
                    Matrix<T> M = getTransitionsForSymbol(i);
                    T[] candidateleftM = (M * (new Matrix<T>(candidateleft))).ColumnVecToList();
                    T[] candidaterightM = (M * (new Matrix<T>(candidateright))).ColumnVecToList();
                    Rule<T> r = new Rule<T>();
                    r.Left = new Vectoroflmelements<T>(candidateleftM);
                    r.Right = new Vectoroflmelements<T>(candidaterightM);
                    todo.Add(r);
                    String wordtoadd = currentword + letter++;
                    wordlist.Add(wordtoadd);
                    wordlist.Add(wordtoadd);


                }
            }
            Console.WriteLine("The following words were explored: ");
            for (int i = 0; i < wordset.Count; ++i) Console.WriteLine(wordset[i] == "" ? "epsilon" : wordset[i]);
            return true;
        }

        public void abstractlyInterpret(T[] vector, T abstraction)
        {
            for (int i = 0; i < vector.Length; ++i) if ((dynamic)vector[i] < abstraction) vector[i] = Zero();
        }



        public static DialogResult InputBox(string title, string promptText, ref string value)
        {
            Form form = new Form();
            Label label = new Label();
            TextBox textBox = new TextBox();
            Button buttonOk = new Button();
            Button buttonCancel = new Button();

            form.Text = title;
            label.Text = promptText;
            textBox.Text = value;

            buttonOk.Text = "OK";
            buttonCancel.Text = "Cancel";
            buttonOk.DialogResult = DialogResult.OK;
            buttonCancel.DialogResult = DialogResult.Cancel;

            label.SetBounds(9, 20, 372, 13);
            textBox.SetBounds(12, 36, 372, 20);
            buttonOk.SetBounds(228, 72, 75, 23);
            buttonCancel.SetBounds(309, 72, 75, 23);

            label.AutoSize = true;
            textBox.Anchor = textBox.Anchor | AnchorStyles.Right;
            buttonOk.Anchor = AnchorStyles.Bottom | AnchorStyles.Right;
            buttonCancel.Anchor = AnchorStyles.Bottom | AnchorStyles.Right;

            form.ClientSize = new System.Drawing.Size(396, 107);
            form.Controls.AddRange(new Control[] { label, textBox, buttonOk, buttonCancel });
            form.ClientSize = new System.Drawing.Size(Math.Max(300, label.Right + 10), form.ClientSize.Height);
            form.FormBorderStyle = FormBorderStyle.FixedDialog;
            form.StartPosition = FormStartPosition.CenterScreen;
            form.MinimizeBox = false;
            form.MaximizeBox = false;
            form.AcceptButton = buttonOk;
            form.CancelButton = buttonCancel;

            DialogResult dialogResult = form.ShowDialog();
            value = textBox.Text;
            return dialogResult;
        }
        public bool bonchiThreshold(T threshold){
            T[] x =new T[1];
            string value = "";
            if (InputBox("Starting vector.", "Please enter the starting vector:", ref value) == DialogResult.OK)
            {
                generateVectorFromString(value, out x);
            }
            Console.WriteLine("Start");
            Console.WriteLine(RuleSystem<T>.SimulationInitialisation(3).ToString());

            Stopwatch stopWatch = new Stopwatch();
            stopWatch.Start();
            bool result = bonchiThreshold(x, threshold);
            stopWatch.Stop();
            long time1 = stopWatch.ElapsedMilliseconds;
            stopWatch.Reset();
            stopWatch.Start();
            bonchiThreshold(x, threshold, true, false); 
            stopWatch.Stop();
            long time2 = stopWatch.ElapsedMilliseconds;
            stopWatch.Reset();
            Console.WriteLine("Algo 1 fertig "+(result?"true":"false")+", Algo 2 startet");
            stopWatch.Start();
            bool result2 = bonchiThresholdUnoptimised(x, threshold);
            stopWatch.Stop();
            long time3 = stopWatch.ElapsedMilliseconds;
            if (result != result2) MessageBox.Show("Hmm");
            MessageBox.Show("Time taken optimised: " + time1 + ", time taken without simulation: " + time2+ ", time taken unoptimised: " + time3);
            Console.WriteLine("Ende");
            return result;
        }
        public bool bonchiVectors()
        {
            T[] x = new T[1];
            T[] y = new T[1];
            string value = "";
            if (InputBox("Vector 1.", "Please enter the first vector to be compared:", ref value) == DialogResult.OK)
            {
                generateVectorFromString(value, out x);
            }
            if (InputBox("Vector 2.", "Please enter the second vector to be compared:", ref value) == DialogResult.OK)
            {
                generateVectorFromString(value, out y);
            }
            return bonchiVectorsRewriting(x, y);
        }

        /// <summary>
        /// An algorithm based on Bonchi's work using rewriting. Ignore for Master thesis
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public bool bonchiThreshold(T[] x, T threshold, bool abstractionOn = true, bool simulation=true)
        {
            string x2 = x[0].ToString(); for (int i=1; i<x.Length; ++i)x2+=", "+x[i].ToString();
            //MessageBox.Show("Entering BonchiThreshold, Threshold " + threshold.ToString()+" Starting Vector: "+x2+", Matrix: \n"+ToString());
            
            List<String> wordlist = new List<String>();
            List<String> wordset = new List<String>();
            lincombcounter = 0;
            Type t = typeof(T);
            RuleSystem<T> R = new RuleSystem<T>();
            if(simulation) R = ThresholdMatrix(this, threshold).getSimulationRelationForWA();
            List<Rule<T>> todo = new List<Rule<T>>();
            Rule<T> initial = new Rule<T>();
            Array.Resize<T>(ref x, x.Length + 1);
            x[x.Length - 1] = Matrix<T>.Zero();
            initial.Left = new Vectoroflmelements<T>(x);
            Vectoroflmelements<T> y = new Vectoroflmelements<T>(x.Length, x.Length - 1, threshold);
            initial.Right = y;
            todo.Add(initial);
            wordlist.Add("");
            wordset.Add("");
            T[] final = getFinalWeightRow();
            while (!(todo.Count() == 0))
            {
                String currentword = wordlist.ElementAt(0);
                wordlist.RemoveAt(0);
                Rule<T> candidate = todo[0];
                todo.Remove(candidate);
                Console.WriteLine("Current word: "+currentword);
                if (R.ApplyAtOnce(candidate.Left) >= candidate.Right) continue;
                //Console.WriteLine("not dropped");
                Rule<T> candidateL = new Rule<T>();
                candidateL.Left = candidate.Left;
                candidateL.Right = candidate.Left + candidate.Right;
                R.Add(candidateL);//no candidateR because of no symmetry!

                /*if (candidate.Left <= R.ApplyAtOnce(candidate.Right)) continue;
                Console.WriteLine("not dropped");
                Rule<T> candidateL = new Rule<T>();
                candidateL.Left = candidate.Right;
                candidateL.Right = candidate.Left + candidate.Right;
                R.Add(candidateL);//no candidateR because of no symmetry!*/


                if (!wordset.Contains(currentword)) wordset.Add(currentword);
                T[] candidateleft = candidate.Left.elements.ToArray<T>();
                T[] candidateright = candidate.Right.elements.ToArray<T>();

                if (!((dynamic)(multiplyVectors(candidateleft, final)) >= threshold))
                {
                    //Console.WriteLine((dynamic)(multiplyVectors(candidateleft, final)));
                    //Console.WriteLine("The following words were explored: ");
                    //for (int i = 0; i < wordset.Count; ++i) Console.WriteLine(wordset[i] == "" ? "epsilon" : wordset[i]);
                    return false;
                }
                char letter = 'a';
                for (int i = 0; i < alphabetsize(); ++i)
                {
                    Matrix<T> M = getTransitionsForSymbol(i, true);
                    T[] candidateleftM = (M * (new Matrix<T>(candidateleft))).ColumnVecToList();
                    T[] candidaterightM = (M * (new Matrix<T>(candidateright))).ColumnVecToList();
                    if (abstractionOn)
                    {
                        //String clm = "(";
                        //for (int k = 0; k < candidateleftM.Length; ++k) clm += candidateleftM[k]+(k==candidateleftM.Length-1?")":", ");
                        abstractlyInterpret(candidateleftM, threshold);
                        //String clm2 = "(";
                        //for (int k = 0; k < candidateleftM.Length; ++k) clm2 += candidateleftM[k] + (k == candidateleftM.Length - 1 ? ")" : ", ");

                        abstractlyInterpret(candidaterightM, threshold);

                        //Console.WriteLine("Abstract Interpretation of " + clm+ " is: "+clm2);
                        
                    }

                    Rule<T> r = new Rule<T>();
                    r.Left = new Vectoroflmelements<T>(candidateleftM);
                    r.Right = new Vectoroflmelements<T>(candidaterightM);
                    todo.Add(r);
                    String wordtoadd = currentword + letter++;
                    wordlist.Add(wordtoadd);
                    //wordlist.Add(wordtoadd);


                }
            }
            //Console.WriteLine("The following words were explored: ");
            //for (int i = 0; i < wordset.Count; ++i) Console.WriteLine(wordset[i] == "" ? "epsilon" : wordset[i]);
            return true;
        }


        /// <summary>
        /// An algorithm based on Bonchi's work using rewriting. Ignore for Master thesis
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public bool bonchiThresholdUnoptimised(T[] x, T threshold, bool abstractionOn=true)
        {
            string x2 = x[0].ToString(); for (int i = 1; i < x.Length; ++i) x2 += ", " + x[i].ToString();
            //MessageBox.Show("Entering BonchiThreshold, Threshold " + threshold.ToString() + " Starting Vector: " + x2 + ", Matrix: \n" + ToString());
            List<String> wordlist = new List<String>();
            List<String> wordset = new List<String>();
            lincombcounter = 0;
            Type t = typeof(T);
            RuleSystem<T> R = new RuleSystem<T>();
            List<Rule<T>> todo = new List<Rule<T>>();
            Rule<T> initial = new Rule<T>();
            Array.Resize<T>(ref x, x.Length + 1);
            x[x.Length - 1] = Matrix<T>.Zero();
            initial.Left = new Vectoroflmelements<T>(x);
            Vectoroflmelements<T> y = new Vectoroflmelements<T>(x.Length, x.Length - 1, threshold);
            initial.Right = y;
            todo.Add(initial);
            wordlist.Add("");
            wordset.Add("");
            while (!(todo.Count() == 0))
            {
                String currentword = wordlist.ElementAt(0);
                wordlist.RemoveAt(0);
                //if(R.Count>0) Console.WriteLine("Größe von R: " + R.Count+", Erstes Element in R hat so viele Einträge: "+R[0].Length);
                Rule<T> candidate = todo[0];
                todo.Remove(candidate);
                Console.WriteLine("Current word: " + currentword);

                if (R.Contains(candidate)) continue;
                //Console.WriteLine("not dropped");
                R.Add(candidate);//no candidateR because of no symmetry!

                /*if (candidate.Left <= R.ApplyAtOnce(candidate.Right)) continue;
                Console.WriteLine("not dropped");
                Rule<T> candidateL = new Rule<T>();
                candidateL.Left = candidate.Right;
                candidateL.Right = candidate.Left + candidate.Right;
                R.Add(candidateL);//no candidateR because of no symmetry!*/


                if (!wordset.Contains(currentword)) wordset.Add(currentword);
                T[] candidateleft = candidate.Left.elements.ToArray<T>();
                T[] candidateright = candidate.Right.elements.ToArray<T>();

                T[] final = getFinalWeightRow();
                if ((dynamic)(multiplyVectors(candidateleft, final)) < threshold)
                {
                    //Console.WriteLine("The following words were explored: ");
                    //for (int i = 0; i < wordset.Count; ++i) Console.WriteLine(wordset[i] == "" ? "epsilon" : wordset[i]);
                    return false;
                }
                char letter = 'a';
                for (int i = 0; i < alphabetsize(); ++i)
                {
                    Matrix<T> M = getTransitionsForSymbol(i, true);
                    T[] candidateleftM = (M * (new Matrix<T>(candidateleft))).ColumnVecToList();
                    T[] candidaterightM = (M * (new Matrix<T>(candidateright))).ColumnVecToList();
                    if (abstractionOn)
                    {
                        abstractlyInterpret(candidateleftM, threshold);
                        abstractlyInterpret(candidaterightM, threshold);
                    }

                    Rule<T> r = new Rule<T>();
                    r.Left = new Vectoroflmelements<T>(candidateleftM);
                    r.Right = new Vectoroflmelements<T>(candidaterightM);
                    todo.Add(r);
                    String wordtoadd = currentword + letter++;
                    wordlist.Add(wordtoadd);
                    //wordlist.Add(wordtoadd);


                }
            }
            //Console.WriteLine("The following words were explored: ");
            //for (int i = 0; i < wordset.Count; ++i) Console.WriteLine(wordset[i] == "" ? "epsilon" : wordset[i]);
            return true;
        }


        /// <summary>
        /// An implementation of a variation of Filipo Bonchi's POPL'13 algorithm for weighted automata. Up to linear combinations and transitivity get optimized, while reflexivity and symmetry get explicitly represented.
        /// In case the semiring is additively idempotent, the optimization from POPL'13 is applicable and is applied provided the flag idempotence is set to true.
        /// Checks, whether two given states of the matrix, interpreted as a transition matrix for a weighted automaton, are language equivalent
        /// </summary>
        /// <param name="x">The index of the first state to be checked</param>
        /// <param name="y">The index of the second state to be checked</param>
        /// <param name="idempotence">Indicates whether the semiring is additively idempotent. In case it is additively idempotent, the optimization from Bonchi's POPL'13 paper is applicable and is being applied.</param>
        /// <returns>1, if x and y have the same language, 0 else.</returns>
        public bool bonchi(int x, int y, bool idempotence = false)
        {
            T[] a = new T[columns()];
            T[] b = new T[columns()];
            for (int i = 0; i < columns(); ++i) a[i] = b[i] = Zero();
            a[x] = b[y] = One();
            return bonchiVectors(a, b, idempotence);
        }

        public static Matrix<T> readMatrix(string filepath)
        {
            //string filePath = @"GanonErgebnisse.csv";
            Type t = typeof(T);
            StreamReader sr = new StreamReader(filepath);
            var lines = new List<string[]>();
            int Row = 0;
            while (!sr.EndOfStream)
            {
                string[] Line = sr.ReadLine().Split(';');
                lines.Add(Line);
                Row++;
            }
            if (lines.Count() == 0) return new Matrix<T>(0, 0);
            Matrix<T> result = new Matrix<T>(lines.Count(), lines[0].Length);
            for (int i = 0; i < lines.Count(); ++i)
                for (int j = 0; j < lines[0].Length; ++j)
                    result.set(i, j, (T)Activator.CreateInstance(t, new object[] { lines[i][j] }));
            return result;
        }

        /*public static Matrix<T> quadratify(Matrix<T> R)
        {
            int dim = (int)Math.Sqrt(R.rows());
            Matrix<T> result = oneMatrix(dim, dim);
            for (int i = 0; i < dim; ++i)
                for (int j = 0; j < dim; ++j)
                    for (int i2 = 0; i2 < dim; ++i2)
                        for (int j2 = 0; j2 < dim; ++j2)
                            result.set(i, j,(dynamic)result.get(i, j) * R.get(i * dim + i2, j * dim + j2));

            return result;

        }*/
        public static Matrix<T> F(Matrix<T> alpha, Matrix<T> beta, Matrix<T> R)
        {
            return elementwiseMult(alpha / (R * transpose(beta)), transpose(beta / transpose(alpha * R)));
            //return quadratify(elementwiseMult(alpha / (R * transpose(beta)), transpose(beta/transpose(alpha*R))));
        }

        Matrix<T> submatrix(int n, int m)
        {
            Matrix<T> result = new Matrix<T>(rows() / m, columns());
            for (int i = 0; i < rows() / m; ++i) result.setRow(i, getRow(i * m + n));
            return result;
        }

        public static Matrix<T> ctsBisim(Matrix<T> alpha, Matrix<T> beta)
        {
            int labelalphabet = alpha.rows() / alpha.columns();
            Matrix<T>[] alphas = new Matrix<T>[labelalphabet];
            Matrix<T>[] betas = new Matrix<T>[labelalphabet];
            for (int i = 0; i < labelalphabet; ++i)
            {
                alphas[i] = alpha.submatrix(i, labelalphabet);
                betas[i] = beta.submatrix(i, labelalphabet);
            }

            Matrix<T> R = oneMatrix(alpha.columns(), beta.columns());
            Matrix<T> oldR;
            do
            {
                //Console.WriteLine(R);
                //Console.ReadLine();
                oldR = R;
                for (int i = 0; i < labelalphabet; ++i)
                {
                    //Console.WriteLine("F for label " + i + ": " + F(alphas[i], betas[i], oldR));
                    //Console.WriteLine("Using alpha i: " + alphas[i] + "and beta i: " + betas[i]);
                    R = elementwiseMult(R, F(alphas[i], betas[i], oldR));
                }

            } while (oldR != R);
            return R;
        }

        public static Matrix<T> zeroMatrix(int i, int j)
        {
            Matrix<T> result = new Matrix<T>(i, j);
            for (int k = 0; (k < i) && (k < j); ++k) result.set(k, k, Zero());
            return result;
        }
        public static Matrix<T> oneMatrix(int i, int j)
        {
            Matrix<T> result = new Matrix<T>(i, j);
            for (int i1 = 0; i1 < i; ++i1)
                for (int j1 = 0; j1 < j; ++j1)
                    result.set(i1, j1, One());
            return result;
        }
        public static Matrix<T> unitMatrix(int i, int j)
        {
            return new Matrix<T>(i, j);
        }

        public bool generateVectorFromString(String input, out T[] result)
        {
            if (input.Contains('\n'))
            {
                MessageBox.Show("Please enter one vector only, no linebreaks.");
                result= new T[1];
                return false;
            }
            if (String.IsNullOrEmpty(input))
            {
                
                MessageBox.Show("Please enter a vector.");
                result= new T[1];
                return false;
            }
            string[] temp1 = System.Text.RegularExpressions.Regex.Split(input, @"\s+");
                           
                //update temp
            int count = 0;
            for (int k = 0; k < temp1.Length; k++)
            {
               if (String.IsNullOrWhiteSpace(temp1[k]))
               {
                   ++count;
               }
            }
            String[] matrixInput = new string [temp1.Length-count];
            int index2 = 0;
            for (int k = 0; k < temp1.Length; k++)
            {              
                if (!String.IsNullOrWhiteSpace(temp1[k]))
                {
                    matrixInput[index2]= temp1[k];
                    index2++;
                }
            }
            int c = matrixInput.Length;
            result = new T[c];


            for (int j = 0; j < c; j++)
            {
                Object obj = Activator.CreateInstance(typeof(T), null);
                Type type = typeof(T);


                try
                {
                    MethodInfo method = type.GetMethod("allowedStrings");
                    Boolean str = (bool)method.Invoke(obj, new Object[] { input[j] });
                    if (str)
                    {
                        try
                        {
                            T oItem = Activator.CreateInstance(typeof(T), new object[]{input[j]}) as T;
                            result[j] = oItem;
                            MessageBox.Show("input["+j+"]: " + input[j]+"result["+j+"]: "+result[j].ToString());

                        }
                        catch
                        {
                            MessageBox.Show("Check your input. Do you set all seperators: [],#  or , correctly?");
                        }

                    }
                    else
                    {
                        MessageBox.Show(input[j] + " is not legal input for this semiring!");
                        return false;
                    }
                }
                // if the user implemented the constructor, in this case no allowed string is generated
                catch
                {
                    // allowedStrings method not implemented, verification should than
                    // be implemented by the user in the string-constructor
                    try
                    {
                        T oItem = Activator.CreateInstance(typeof(T), new object[]{matrixInput[j]}) as T;
                        result[j] = oItem;

                    }
                    catch
                    {
                        MessageBox.Show("Check your input. Do you set all seperators: [] and , correctly?");
                    }

                }
            }



            return true;
        }

        public RuleSystem<T> getSimulationRelationForWA()
        {
            Console.WriteLine(this);
            T[] final = getFinalWeightRow();
            RuleSystem<T> result = RuleSystem<T>.SimulationInitialisation(this.columns());
            int n;
            for (int ruleindex = 0; ruleindex < result.n(); ++ruleindex)
            {
                Rule<T> currentRule = result.Get(ruleindex);
                if (!((dynamic)currentRule.Left * final >= (dynamic)currentRule.Right * final))
                {
                    result.Delete(ruleindex--);
                }
            }
            do
            {
                n = result.n();
                for (int ruleindex = 0; ruleindex < result.n(); ++ruleindex)
                {
                    Rule<T> currentRule = result.Get(ruleindex);
                    for (int i = 0; i < alphabetsize(); ++i)
                    {
                        Matrix<T> M = getTransitionsForSymbol(i, false);

                        Vectoroflmelements<T> v2 = new Vectoroflmelements<T>((M * (new Matrix<T>(currentRule.Left))).ColumnVecToList());
                        Vectoroflmelements<T> v1 = new Vectoroflmelements<T>((M * (new Matrix<T>(currentRule.Right))).ColumnVecToList());
                        //Vectoroflmelements<T> s1 = new Vectoroflmelements<T>(v1.Length(), 0, Zero());
                        Vectoroflmelements<T> s1 = new Vectoroflmelements<T>(v1.Length(), 0, Zero());
                        for (int rulectr = 0; rulectr < result.n(); ++rulectr)
                        {
                            //Console.WriteLine(rulectr + ", " + result.n());
                            Rule<T> rule = result.Get(rulectr);
                            T multiplicand = (v2 / rule.Left);
                            //multiplicand = ((dynamic)multiplicand).infimum(v2 / rule.Right);
                            //s1 = s1 + rule.Left * multiplicand;
                            s1 = s1 + rule.Right * multiplicand;
                        }
                        if (!(v1<=s1))
                        {
                            result.Delete(ruleindex--);
                            i = alphabetsize();//To skip check of further letters and continue with the next rule
                        }
                    }

                }
            } while (n != result.n());

            for (int i = 0; i < result.n(); ++i)
                if (result.Get(i).Left == result.Get(i).Right) result.Delete(i--); else result.Get(i).Right = result.Get(i).Left + result.Get(i).Right;

            Console.WriteLine("Resulting Simulation Relation: ");
            Console.WriteLine(result);
            return result;
        }
        Matrix<T> ThresholdMatrix(Matrix<T> M, T threshold)
        {
            Matrix<T> result = new Matrix<T>(M.rows()+alphabetsize(), M.columns()+1);
            for (int i = 0; i < M.rows()-1; ++i)
                for (int j = 0; j < M.columns(); ++j)
                    result.set(i, j, M.get(i, j));
            for (int i = 0; i < M.rows()-1; ++i) result.set(i, M.columns(), Zero());
            for (int j = 0; j < alphabetsize(); ++j)
            {
                for (int i = 0; i < M.columns(); ++i) result.set(M.rows() - 1+j, i, Zero());
                result.set(M.rows() - 1+j, M.columns(), One());
            }
            for (int i = 0; i < M.columns(); ++i) result.set(M.rows() + alphabetsize() - 1, i, M.get(M.rows() - 1, i));
            result.set(M.rows()+alphabetsize()-1, M.columns(), threshold);
            return result;
        }
    }
}


