﻿using System;
using System.Collections.Generic;
using System.Linq;

using System.Reflection;
using System.IO;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Windows.Forms;
using Analysis_Part;


namespace Generator_2._0
{
    /// <summary>
    /// Parent class for code generation. Includes all necessary methods for building code based on CodeDom.
    /// https://msdn.microsoft.com/de-de/library/y2k85ax6%28v=VS.110%29.aspx:
    /// "You can design a program that builds a source code model using CodeDOM elements to assemble an object graph."
    /// All elements build on the abstract class: CodeStatement
    ///
    /// syntax parts:
    /// 1. loop constructions(single or mutli codestatments)
    /// 2. If blocks (single or mutli codestatments)
    /// 3. if else blocks (single or mutli codestatments)
    /// Semantic parts:
    /// 4. Constructor
    /// 5. Member
    /// 6. Methods (single or mutli codestatments)
    /// 7. single or mutli codestatments
    /// 8. Direct user Input is handled by CodeSnippet, which is not generic in the sense of multiple language suport
    /// 
    ///This class is directly used for "arbitray semirings" and by the child-class for "L-Monoids".
    /// </summary>
    class Generator
    {
        /// <summary>
        /// Define the compile unit to use for code generation. 
        /// Define the new Class
        /// Define of namespace
        /// </summary>
        protected CodeCompileUnit targetUnit;
        protected CodeTypeDeclaration targetClass;
        protected CodeNamespace samples;

        /// <summary>
        /// The name of the file to contain the source code.
        /// </summary>
        protected string outputFileName;
        /// <summary>
        /// The name of the new class.
        /// </summary>
        protected string classname;

        /// <summary>
        /// This List is initialized corresponding the user selection.
        /// The type information is needed for:
        /// Returntype-
        /// Parameter-
        /// temporary variable-
        /// construction
        /// </summary>
        protected List<Type> basetypes;


        /// <summary>
        /// The Generator Constructor:
        /// Includes all information to build the new class:
        /// - Imports
        /// - BaseTypes
        /// - Namespace
        /// </summary>
        public Generator(String classname, string t)
        {
            this.classname = classname;
            outputFileName = classname + ".cs";
            targetUnit = new CodeCompileUnit();
            samples = new CodeNamespace("Analysis_Part");
            samples.Imports.Add(new CodeNamespaceImport("System"));
            samples.Imports.Add(new CodeNamespaceImport("System.Collections.Generic"));
            samples.Imports.Add(new CodeNamespaceImport("System.Linq"));
            samples.Imports.Add(new CodeNamespaceImport("System.Text"));
            samples.Imports.Add(new CodeNamespaceImport("Microsoft.VisualBasic"));


            targetClass = new CodeTypeDeclaration(classname);
            // Use attributes to mark the class as serializable.
            CodeAttributeDeclaration codeAttrDecl =
                new CodeAttributeDeclaration("System.Serializable");
            targetClass.CustomAttributes.Add(codeAttrDecl);


            if (t != "") targetClass.BaseTypes.Add(new CodeTypeReference(t));
            else targetClass.BaseTypes.Add(new CodeTypeReference("Parser"));// String-Constructor needs parsing-methods
            targetClass.IsClass = true;
            targetClass.TypeAttributes =
                TypeAttributes.Public;
            samples.Types.Add(targetClass);
            targetUnit.Namespaces.Add(samples);
        }


        /// <summary>
        /// Attributes can contain finite SR like FiniteLattices or Modulo
        /// Information must be seperated for Codegeneration
        /// </summary>
        /// <param name="attributes">List of Semirings or primitive datatypes</param>
        /// <returns>Modulo and FiniteLattice information</returns>
        public Dictionary<int, string> finiteSRInfo(List<string> attributes)
        {
            Dictionary<int, string> finite_SR = new Dictionary<int, string>();
            for (int i = 0; i < attributes.Count; i++)
            {
                if (attributes[i].Contains("-"))
                {
                    try
                    {
                        string[] split = attributes[i].Split('-');
                        if (split[0] == "Modulo")
                        {
                            attributes[i] = split[0];
                            finite_SR.Add(i, split[1]);
                        }
                        if (split[1] == "FiniteLattice")
                        {
                            attributes[i] = split[1];
                            finite_SR.Add(i, split[0]);
                        }

                        if (split[1] == "LatticeBasedOnBDD")
                        {
                            attributes[i] = split[1];
                            finite_SR.Add(i, split[0]);
                        }

                    }
                        // just ignore if some other classname contains "-"
                    catch (Exception outOfBounce){ }

                }
            }

            return finite_SR;
        }
        public bool generateSR(List<string> attributes, Dictionary<string, string> methodcode, Dictionary<string, string> optional, Dictionary<string, List<string>> info)
        {
            // prepare info under consideration of Modulo and FiniteLattice
            Dictionary<int, string> finite_SR =  finiteSRInfo(attributes);

            // attributes of the new class:
            generateAttributes(attributes);

            if (basetypes.Count != attributes.Count) return false;

            // empty constructor, is used by WindowsFormsApplication while creating a Matrix<T> by the corresponding data model
            // Object obj = Activator.CreateInstance(typeof(T), null) to enable checking the user input before creating the Matrix<T>
            // (Creates an instance of the specified type using the constructor that best matches the specified parameters.)
            CodeConstructor constructor = new CodeConstructor();
            constructor.Attributes = MemberAttributes.Public;
            targetClass.Members.Add(constructor);

            // necessary methods and constructors of the new class:
            /*
             *extraxtMethodStatements builds the AST calling the corresponding behaviour generating methode
             *addNecessaryOperators builds the function-definition (a syntax-helping method)
             * 
             */
            foreach (KeyValuePair<string, string> entry in methodcode)
            {
                switch (entry.Key)
                {
                    case Constants.LINCOMBI:
                        if (entry.Value.Contains("Gauss")) getGauss(attributes, targetClass);
                        if (entry.Value.Contains("Largest")) targetClass.Members.Add(GeneratorLMonoid.getLargestSolution(attributes, targetClass));
                        else if (entry.Value == Constants.SB)
                        {
                            CodeStatement[] lincom = extractMethodsStatements(entry.Value, "LinearCombination", attributes, finite_SR);
                            targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "find linearcombination", "LinearCombination", "LinearCombination<" + targetClass.Name + ">", lincom));
                        }
                        break;
                    case Constants.C1:
                        if (entry.Value.Contains("Componentwise")) targetClass.Members.Add(addConstructorNewClass(attributes, targetClass.Name));
                        else targetClass.Members.Add(addConstructor(targetClass.Name, entry.Key, extractCode(entry.Value), attributes));
                        break;
                    case Constants.C2:
                        if (entry.Value.Contains("Componentwise")) targetClass.Members.Add(addConstructorByAttri(attributes));
                        else targetClass.Members.Add(addConstructor(targetClass.Name, entry.Key, extractCode(entry.Value), attributes));
                        break;
                    case Constants.C3:
                        if (entry.Value.Contains("Componentwise")) targetClass.Members.Add(generateStringConstructor(attributes, finite_SR));
                        else targetClass.Members.Add(addConstructor(targetClass.Name, entry.Key, extractCode(entry.Value), attributes));
                        break;
                    case Constants.Z:
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "Yields the Zero Element:", "Zero", targetClass.Name, getMethod(attributes, "Zero()", finite_SR)));
                        break;
                    case Constants.O:
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "Yields the One Element:", "One", targetClass.Name, getMethod(attributes, "One()",finite_SR)));
                        break;
                    case Constants.ADD:
                        CodeStatement[] plus = extractMethodsStatements(entry.Value, "+", attributes, finite_SR);
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : +", "operator +", targetClass.Name, plus));
                        break;
                    case Constants.MUL:
                        CodeStatement[] mul = extractMethodsStatements(entry.Value, "*", attributes, finite_SR);
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : *", "operator *", targetClass.Name, mul));
                        break;
                    case Constants.SUB:
                        CodeStatement[] minus = extractMethodsStatements(entry.Value, "-", attributes, finite_SR);
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : -", "operator -", targetClass.Name, minus));
                        break;
                    case Constants.DIV:
                        // no automatic-support for componentwise division
                        CodeStatement[] div = extractMethodsStatements(entry.Value, "/", attributes, finite_SR);
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : /", "operator /", targetClass.Name, div));
                        break;
                    case Constants.EQUAL:
                        CodeStatement[] eq = extractMethodsStatements(entry.Value, "==", attributes, finite_SR);
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : ==", "operator ==", "System.Boolean", eq));
                        break;
                    case Constants.GREATER:
                        CodeStatement[] greater = extractMethodsStatements(entry.Value, ">", attributes, finite_SR);
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : >", "operator >", "System.Boolean", greater));
                        //< automatic generation
                        CodeStatement[] lassThan = new CodeStatement[1];
                        lassThan[0] = new CodeExpressionStatement(new CodeVariableReferenceExpression("return ((b>a) && !(a==b))"));
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : <", "operator <", "System.Boolean", lassThan));
                        lassThan[0] = new CodeExpressionStatement(new CodeVariableReferenceExpression("return ((b>a) || (a==b))"));
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : <=", "operator <=", "System.Boolean", lassThan));
                        lassThan[0] = new CodeExpressionStatement(new CodeVariableReferenceExpression("return ((a>b) || (a==b))"));
                        targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : >=", "operator >=", "System.Boolean", lassThan));
                        break;
                }
   
            }
           

            // add the (!=) operator, based on negation of (==)
            CodeStatement[] st = new CodeStatement[1];
            st[0] = new CodeExpressionStatement(new CodeVariableReferenceExpression("return !(a==b)"));
            targetClass.Members.Add(addNecessaryOperators(targetClass.Name, "necessary operator : !=", "operator !=", "System.Boolean", st));

            //targetClass.Members.Add(minusT(targetClass.Name, attributes)); // was for minus operator
            targetClass.Members.Add(getToString(attributes));

            // generate necessary methods while working with modules based on datastructures like arrays:
            // do not add multiply times
            List<string> helpingMethods = new List<string>();
            for (int i = 0; i < basetypes.Count(); i++)
            {
                if (basetypes[i].IsArray)
                {
                    if (!helpingMethods.Contains(basetypes[i].Name)) helpingMethods.Add(basetypes[i].Name); 
                }
            }

            for (int i = 0; i < helpingMethods.Count(); i++)
            {

                // before testing:        if (methodcode.Keys.Contains(Constants.SUB)) targetClass.Members.Add(minus(helpingMethods[i].Substring(0, attributes[i].Length - 2)));
                if (methodcode.Keys.Contains(Constants.SUB)) targetClass.Members.Add(minus(helpingMethods[i].Substring(0, helpingMethods[i].Length - 2)));
                targetClass.Members.Add(ifZeroCut(helpingMethods[i].Substring(0, helpingMethods[i].Length - 2)));
                targetClass.Members.Add(equals(helpingMethods[i]));

            }

            // a user has the option to add other methdos to his Module or Semiring
            // this is managed by: getOptionalMethod
            // info 1. Parameter 2. returntype
            foreach (KeyValuePair<string, string> entry in optional)
            {
                targetClass.Members.Add(getOptionalMethod(entry.Key, entry.Value, info[entry.Key][0], info[entry.Key][1]));
            }

            /* There is a need to check, if the userinput for an arbitrary semiring satisfies the requirements
               this method is invoked  before creating a Matrix<T> in the Model<T> of the MVC
               seperated from the constructor for better overview while constructing the CodeDOM-Diagram
               check if string constructor is implemented by the user! otherwise construct allowed string componentwise:
            */
            if (methodcode[Constants.C3] != Constants.SB) targetClass.Members.Add(addAllowedString(attributes,finite_SR));
            return true;
        }





        /*
         * Basic methods for managing construction of CodeDom Diagramm
         * Members:
         * 
         * Attributes ,Construtors & Methods
         * 
         */ 

        /// <summary>
        /// Depending on the seleceted attributes by the user using the GUI of Generator, the class attributes are generated.
        /// </summary>
        /// <param name="member"></param>
        protected void generateAttributes(List<string> member)
        {
            basetypes = new List<Type>();
            for (int i = 0; i < member.Count(); i++)
            {
                CodeMemberField ele = new CodeMemberField();
                // getFieldInfo is needed for string constructor, using special bindingflags enables getting all private Fields
                ele.Attributes = MemberAttributes.Private;
                ele.Name = "mem" + i;

                string help = member[i];
                Type t;
                switch (help)
                {
                    case "int":
                        t = typeof(System.Int32);
                        member[i] = "System.Int32";
                        break;
                    case "uint":
                        t = typeof(System.UInt32);
                        member[i] = "System.UInt32";
                        break;
                    case "double":
                        t = typeof(System.Double);
                        member[i] = "System.Double";
                        break;
                    case "double[]":
                        t = typeof(System.Double[]);
                        member[i] = "System.Double[]";
              
                        break;
                    case "int[]":
                        t = typeof(System.Int32[]);
                        member[i] = "System.Int32[]";
              
                        break;

                    default:
                        t = getType(help);

                        break;
                }

                try
                {
                    ele.Type = new CodeTypeReference(t);
                    targetClass.Members.Add(ele);
                    basetypes.Add(t);
                }
                catch
                {
                    MessageBox.Show("Basetype: "+ help +" is wrong! It was deleted!");
                }
            }
        }

        /// <summary>
        /// Get type method including all types of AlgorithmB Application.
        /// This Information is needed for the whole generator-programm
        /// </summary>
        /// <param name="attr"></param>
        /// <returns></returns>
        public static Type getType(string attr)
        {
           
            Type t;
            t = Type.GetType(attr);


            if (t != null)
            {
                return t;
            }
            else
            {
                foreach (Assembly asm in AppDomain.CurrentDomain.GetAssemblies())
                {
                    t = asm.GetType(attr);
                    if (t != null)
                        return t;
                }
            }


            string targetPath = Directory.GetParent(Directory.GetParent(Directory.GetParent(Directory.GetCurrentDirectory()).ToString()).ToString()).ToString() + "\\Analysis_Part\\Analysis_Part\\";
            byte[] readAllBytes = File.ReadAllBytes(targetPath + "bin\\debug\\Analysis_Part.exe");
            Assembly assembly = Assembly.Load(readAllBytes);

            if (attr.Contains("Polynom"))
            {
                Type generic = typeof(Polynom<>);
                string[] argument = attr.Split('<');
                if (argument[1].Substring(0, argument[1].Length - 1) != "double")
                {
                    string arg = "Analysis_Part." + argument[1].Substring(0, argument[1].Length - 1);
                    Type[] typeArgs = { assembly.GetType(arg) };
                    t = generic.MakeGenericType(typeArgs);
                }
                else t = generic.MakeGenericType(typeof(System.Double));
            }

            //also is workin for Semiring[] correctly
            else if (t == null) t = assembly.GetType("Analysis_Part." + attr);

            return t;
        }


        /// <summary>
        /// enables direct user input in c# code for constructors
        /// </summary>
        /// <param name="targetClass"></param>
        /// <param name="c"></param>
        /// <param name="st"></param>
        /// <param name="attributes"></param>
        /// <returns></returns>
        public  CodeConstructor addConstructor(string targetClass, string c, CodeStatement st, List<string> attributes)
        {
            CodeConstructor constructor = new CodeConstructor();
            constructor.Attributes = MemberAttributes.Public;
            if (c == Constants.C1) constructor.Parameters.Add(new CodeParameterDeclarationExpression(targetClass, "x"));
            if (c == Constants.C2)
            {
                for (int i = 0; i < attributes.Count(); i++)
                {
                    constructor.Parameters.Add(new CodeParameterDeclarationExpression(attributes[i], "a" + i));
                    CodeAssignStatement as1 = new CodeAssignStatement(new CodeVariableReferenceExpression("this.mem" + i), new CodeVariableReferenceExpression("a" + i));  
                    constructor.Statements.Add(as1);
                    if (basetypes[i].IsArray)
                    {
                        constructor.Statements.Add(new CodeVariableReferenceExpression("this.mem" + i + "= ifZeroCut(this.mem" + i + ")"));
                    }
                }

            }
            if (c == Constants.C3) constructor.Parameters.Add(new CodeParameterDeclarationExpression(typeof(System.String), "str"));

            constructor.Statements.Add(st);
            return constructor;
        }

        /// <summary>
        /// Template for componentwise-Constructor with his own type as single parameter
        /// </summary>
        /// <param name="t"></param>
        /// <param name="targetClass"></param>
        /// <returns></returns>
        public static CodeConstructor addConstructorNewClass(List<string> attributes, string targetClass)
        {
            CodeConstructor constructor = new CodeConstructor();
            constructor.Attributes = MemberAttributes.Public;
            constructor.Parameters.Add(new CodeParameterDeclarationExpression(targetClass, "x"));
            for (int i = 0; i < attributes.Count(); i++)
            {
                CodeVariableReferenceExpression con1;
                if (attributes[i] == "System.Int32" || attributes[i] == "System.Double")
                {
                    con1 = new CodeVariableReferenceExpression("this.mem" + i + "= x.mem" + i);
                    constructor.Statements.Add(con1);
                }
     
                if (!attributes[i].Contains("Int") && !attributes[i].Contains("Double") &&  !attributes[i].Contains("["))
                {
                    // deep copy:
                    con1 = new CodeVariableReferenceExpression("this.mem" + i + "= new " + attributes[i] + "( x.mem" + i + ")");
                    constructor.Statements.Add(con1);
                }

                if (attributes[i] == "System.Double[]" || attributes[i] == "System.Int32[]")
                {
                    string type ="int";
                    if(attributes[i] == "System.Double[]")type="double";
                    // deep copy:
                    constructor.Statements.Add(new CodeVariableReferenceExpression("this.mem" + i + "= new " + type + "[x.mem" + i + ".Length]"));
                    con1 = new CodeVariableReferenceExpression("this.mem" + i + "[i] = x.mem" + i + "[i]");
                    constructor.Statements.Add(addForLoop("x.mem"+i+".Length","i",con1));
                    constructor.Statements.Add(new CodeVariableReferenceExpression("this.mem" + i + "= ifZeroCut(this.mem" + i + ")"));
                }
                if (!attributes[i].Contains("Int") && !attributes[i].Contains("Double") && attributes[i].Contains("["))
                {
                    // deep copy:
                    constructor.Statements.Add(new CodeVariableReferenceExpression("this.mem" + i + "= new " + attributes[i].Substring(0,attributes[i].Length-2) + "[x.mem" + i + ".Length]"));
                    con1 = new CodeVariableReferenceExpression("this.mem" + i + "[i] = new " + attributes[i].Substring(0,attributes[i].Length-2)+"(x.mem" + i + "[i])");
                    constructor.Statements.Add(addForLoop("x.mem" + i + ".Length", "i", con1));
                    constructor.Statements.Add(new CodeVariableReferenceExpression("this.mem" + i + "= ifZeroCut(this.mem" + i + ")"));
                }

       
            }
            return constructor;
        }

        /// <summary>
        /// Template for componentwise-Constructor with the attributes as parameters
        /// </summary>
        /// <param name="t"></param>
        /// <param name="targetClass"></param>
        /// <returns></returns>
        public static CodeConstructor addConstructorByAttri(List<string> attributes)
        {
            CodeConstructor constructor = new CodeConstructor();
            constructor.Attributes = MemberAttributes.Public;
            for (int i = 0; i < attributes.Count(); i++)
            {
                constructor.Parameters.Add(new CodeParameterDeclarationExpression(attributes[i], "a"+i ));
                CodeVariableReferenceExpression con1;
                if (attributes[i] == "System.Int32" || attributes[i] == "System.Double")
                {
                    con1 = new CodeVariableReferenceExpression("this.mem" + i + "= a" + i);
                    constructor.Statements.Add(con1);
                }

                if (!attributes[i].Contains("Int") && !attributes[i].Contains("Double") && !attributes[i].Contains("["))
                {
                    // deep copy:
                    con1 = new CodeVariableReferenceExpression("this.mem" + i + "= new " + attributes[i] + "( a" + i + ")");
                    constructor.Statements.Add(con1);
                }

                if (attributes[i] == "System.Double[]" || attributes[i] == "System.Int32[]")
                {
                    string type = "int";
                    if (attributes[i] == "System.Double[]") type = "double";
                    // deep copy:
                    constructor.Statements.Add(new CodeVariableReferenceExpression("this.mem" + i + "= new " + type + "[a" + i + ".Length]"));
                    con1 = new CodeVariableReferenceExpression("this.mem" + i + "[i] =  a" + i + "[i]");
                    constructor.Statements.Add(addForLoop("a" + i + ".Length", "i", new CodeExpressionStatement(con1), CodeBinaryOperatorType.Add));
                }

                if (!attributes[i].Contains("Int") && !attributes[i].Contains("Double") && attributes[i].Contains("["))
                {
                    // deep copy:
                    constructor.Statements.Add(new CodeVariableReferenceExpression("this.mem" + i + "= new " + attributes[i].Substring(0, attributes[i].Length - 2) + "[a" + i + ".Length]"));
                    con1 = new CodeVariableReferenceExpression("this.mem" + i + "[i]= new " + attributes[i].Substring(0, attributes[i].Length - 2) + "( a" + i + "[i])");
                    constructor.Statements.Add(addForLoop("a" + i + ".Length", "i", con1));
                }


                if (attributes[i].Contains("[")) constructor.Statements.Add(new CodeVariableReferenceExpression("this.mem" + i.ToString() + "= ifZeroCut(this.mem" + i.ToString() + ")"));
            }
            return constructor;
        }

        /// <summary>
        /// automatic string constructor, which depends on the selected attributes
        /// working componentwise, this means this constructor-template generates constructors,
        /// which allways expects input depending on primtitive datatypes, arrays or
        /// the semiring of complex numbers (called finale datatypes)
        /// otherwise the input will be passed to the underlying constructor
        /// 
        /// </summary>
        /// <param name="member">attributes/fields of the SR</param>
        /// <param name="finite_SR">Info of finite SR</param>
        /// <returns>CodeConstructor</returns>
        public CodeConstructor generateStringConstructor(List<string> member, Dictionary<int, string> finite_SR)
        {
            // the constructor attributes and parameters:
            CodeConstructor constructor = new CodeConstructor();
            constructor.Attributes = MemberAttributes.Public;
            constructor.Parameters.Add(new CodeParameterDeclarationExpression(typeof(System.String), "str"));


            // generate typeinfo for finite semirings
            // get the basetype for input:
            // convert attributes into array of fieldType
            Type[] obj = new Type[member.Count()];
            for (int i = 0; i < member.Count(); i++)
            {
                obj[i] = getType(member[i]);
            }
            List<Type> result = new List<Type>();
            getBaseTypeOrder(obj, 0, ref result);
            createInputInfo(result,finite_SR);
       
            // position of arrays and other basetypes
            List<int> array_info = new List<int>();
            for (int i = 0; i < result.Count(); i++)
            {
                if (result[i].IsArray) array_info.Add(i);
                if (result[i].Name == "FiniteLattice") array_info.Add(i);
            }
            CodeVariableReferenceExpression infor = new CodeVariableReferenceExpression("info");
            CodeArrayCreateExpression cai = new CodeArrayCreateExpression(typeof(int), array_info.Count());
            CodeVariableDeclarationStatement arrayinfo = new CodeVariableDeclarationStatement("System.Int32[]", "info", cai);
            constructor.Statements.Add(arrayinfo);
            for (int i = 0; i < array_info.Count(); i++)
            {
                CodeAssignStatement ass_st = new CodeAssignStatement(new CodeArrayIndexerExpression(infor, new CodePrimitiveExpression(i)), new CodePrimitiveExpression(array_info[i]));
                constructor.Statements.Add(ass_st);
            }


                constructor.Statements.Add(new CodeVariableDeclarationStatement(typeof(string[]), "split", new CodeMethodInvokeExpression(null, "getSplit", Constants.str, new CodeVariableReferenceExpression("info"), new CodePrimitiveExpression(result.Count))));
      
                int start = 0;
                for (int i = 0; i < member.Count(); i++)
                {
                    //Check for MODULO and FINITELATTICE UPDATES
                    if (finite_SR.Keys.Contains(i))
                    {
                        string loadType = "";
                        if (member[i] == "Modulo") loadType = "Modulo.modulo =" + finite_SR[i];
                        if (member[i] == "FiniteLattice") loadType = "PreorderModel.getInstance.loadPreorder(\"" + finite_SR[i] + "\")";
                        if (member[i] == "LatticeBasedOnBDD") loadType = "BDDModel.getInstance.loadBDD(\"" + finite_SR[i] + "\")";
                    CodeExpressionStatement lt = new CodeExpressionStatement(new CodeVariableReferenceExpression(loadType));
                        constructor.Statements.Add(lt);
                    }


                    int field = 0;
                    Type t = Type.GetType(member[i]);
                    if (t == null) t = getType(member[i]);
                    //Type t = basetypes[i];
                    FieldInfo[] info = t.GetFields(BindingFlags.NonPublic | BindingFlags.Instance);
                    if (info.Count() > 0 && basetypes[i].Name != "Complex" && basetypes[i].Name != "FiniteLattice") field += getNumberofFields(info, field);
                    else field += 1;

                    CodeExpression ass = cunstructConcatenation(member[i],t,start,field,true);
                 
                    if (t == typeof(uint)) ass = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToUInt32", new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)));
                    if (t == typeof(int)) ass = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToInt32", new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)));
                    if (t == typeof(double)) ass = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToDouble", new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)));
                    if (member[i] == "Complex") ass = new CodeObjectCreateExpression(new CodeTypeReference(member[i]), new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)));
                    // extra fo arrays:

                    CodeStatement st = new CodeAssignStatement(new CodeVariableReferenceExpression("mem" + i), ass);
                    if (member[i].Contains("[") ||member[i].Contains("FiniteLattice"))
                    {
                        
                     
                        //mem+i = new member[i] [values.Lenght]
                        // for ( int i=0; i< values.Length;i++)
                        //   {
                        //      memi[i]= new SR (value[i]+","+value[i+1]...);
                        //      i=i+field;
                        //   }
                        if (!member[i].Contains("Double") && !member[i].Contains("Int") && !member[i].Contains("UInt"))
                        {
                            string sr_type = member[i].Substring(0, member[i].Length - 2);
                            if (member[i] == "FiniteLattice")
                            {
                                sr_type = "FiniteLattice";
                                st = new CodeExpressionStatement(new CodeVariableReferenceExpression("mem"+i +"= new FiniteLattice(split["+i+"].Substring(1,split["+i+"].Length-2))"));
                            }
                            else
                            {
                                //call the getSplitArrayInput method to get a List<string> values
                                CodeVariableDeclarationStatement values = new CodeVariableDeclarationStatement("List<string>", "values" + i, new CodeObjectCreateExpression("List<string>"));
                                //getArraySplit
                                constructor.Statements.Add(values);
                                //getArraySplit(split[1].Substring(1,split[1].Length-2), ref values1);
                                CodeBinaryOperatorExpression substring_len = new CodeBinaryOperatorExpression(new CodeFieldReferenceExpression(new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)), "Length"), CodeBinaryOperatorType.Subtract, new CodePrimitiveExpression(2));
                                CodeMethodInvokeExpression substring = new CodeMethodInvokeExpression(new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)), "Substring", Constants.one, substring_len);
                                constructor.Statements.Add(new CodeMethodInvokeExpression(null, "getArraySplit", substring, new CodeVariableReferenceExpression("ref values" + i)));

                                // inside for loop, < values.Length
                                string concat = "values" + i + "[i]";
                                //getNumber of final fields for array type:
                                int finaltypes = getNumberofFields(getType(member[i].Substring(0, member[i].Length - 2)).GetFields(BindingFlags.NonPublic | BindingFlags.Instance), 0);
                                CodeFieldReferenceExpression val_len = new CodeFieldReferenceExpression(new CodeVariableReferenceExpression("values" + i), "Count");
                                CodeBinaryOperatorExpression length = new CodeBinaryOperatorExpression(val_len, CodeBinaryOperatorType.Divide, new CodePrimitiveExpression(finaltypes));
                                constructor.Statements.Add(new CodeAssignStatement(new CodeVariableReferenceExpression("mem" + i), new CodeArrayCreateExpression(member[i].Substring(0, member[i].Length - 2), length)));
                                for (int j = 1; j < finaltypes; ++j)
                                {
                                    // not generic, a code snippet
                                    concat += "+" + "','" + "+" + "values" + i + "[i+" + j + "]";

                                }
                                CodeObjectCreateExpression insideLoop = new CodeObjectCreateExpression(new CodeTypeReference(member[i].Substring(0, member[i].Length - 2)), new CodeVariableReferenceExpression(concat));
                                CodeAssignStatement inside2 = new CodeAssignStatement(new CodeVariableReferenceExpression("i"), new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression("i"), CodeBinaryOperatorType.Add, new CodePrimitiveExpression(finaltypes - 1)));
                                CodeAssignStatement inside1 = new CodeAssignStatement(new CodeArrayIndexerExpression(new CodeVariableReferenceExpression("mem" + i), new CodeVariableReferenceExpression("index" + i)), insideLoop);
                                CodeStatement[] inside = new CodeStatement[3];
                                inside[0] = inside1;
                                inside[1] = inside2;
                                inside[2] = new CodeAssignStatement(new CodeVariableReferenceExpression("index" + i), new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression("index" + i), CodeBinaryOperatorType.Add, new CodePrimitiveExpression(1)));
                                constructor.Statements.Add(new CodeVariableDeclarationStatement(new CodeTypeReference("System.Int32"), "index" + i, new CodePrimitiveExpression(0)));
                                st = addForLoopStatements(new CodeFieldReferenceExpression(new CodeVariableReferenceExpression("values" + i), "Count"), "i", new CodePrimitiveExpression(0), inside);


                            }

                        }
                        else
                        {
                            //split[member].Length
                            CodeFieldReferenceExpression length = new CodeFieldReferenceExpression(new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)), "Length");
                            CodeVariableDeclarationStatement array = new CodeVariableDeclarationStatement(typeof(string), "split_array" + i, new CodeMethodInvokeExpression(new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)), "Substring", new CodeExpression[] { new CodePrimitiveExpression(1), new CodeBinaryOperatorExpression(length, CodeBinaryOperatorType.Subtract, new CodePrimitiveExpression(2)) }));
                            constructor.Statements.Add(array);
                            // [a,b,c] -> a,b,c
                            CodeVariableDeclarationStatement array_split = new CodeVariableDeclarationStatement(typeof(string[]), "values" + i, new CodeMethodInvokeExpression(new CodeVariableReferenceExpression("split_array" + i), "Split", new CodePrimitiveExpression(',')));
                            constructor.Statements.Add(array_split);
                            CodeVariableReferenceExpression mem = new CodeVariableReferenceExpression("mem" + i);
                            CodeFieldReferenceExpression length2 = new CodeFieldReferenceExpression(new CodeVariableReferenceExpression("values" + i), "Length");
                            CodeAssignStatement create = new CodeAssignStatement(mem, new CodeArrayCreateExpression(member[i], length2));
                            //mem+i = new member[i] [values.Lenght]
                            // for ( int i=0; i< values.Length;i++)
                            //   {
                            //      memi[i]= Convert.ToDouble(value[i]);
                            //   }
                            constructor.Statements.Add(create);
                            if (member[i].Contains("Double"))
                            {

                                CodeMethodInvokeExpression left = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToDouble", new CodeArrayIndexerExpression(new CodeVariableReferenceExpression("values" + i), new CodeVariableReferenceExpression("i")));
                                CodeAssignStatement inside = new CodeAssignStatement(new CodeArrayIndexerExpression(mem, new CodeVariableReferenceExpression("i")), left);
                                st = addForLoop(new CodeFieldReferenceExpression(new CodeVariableReferenceExpression("values" + i), "Length"), "i", inside);
                            }
                            if (member[i].Contains("Int32"))
                            {
                                CodeMethodInvokeExpression left = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToInt32", new CodeArrayIndexerExpression(new CodeVariableReferenceExpression("values" + i), new CodeVariableReferenceExpression("i")));
                                CodeAssignStatement inside = new CodeAssignStatement(new CodeArrayIndexerExpression(mem, new CodeVariableReferenceExpression("i")), left);
                                st = addForLoop(new CodeFieldReferenceExpression(new CodeVariableReferenceExpression("values" + i), "Length"), "i", inside);
                            }

                            if (member[i].Contains("UInt32"))
                            {
                                CodeMethodInvokeExpression left = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToUInt32", new CodeArrayIndexerExpression(new CodeVariableReferenceExpression("values" + i), new CodeVariableReferenceExpression("i")));
                                CodeAssignStatement inside = new CodeAssignStatement(new CodeArrayIndexerExpression(mem, new CodeVariableReferenceExpression("i")), left);
                                st = addForLoop(new CodeFieldReferenceExpression(new CodeVariableReferenceExpression("values" + i), "Length"), "i", inside);
                            }


                        }

                    }

         

                    start += field;
                    constructor.Statements.Add(st);
                }

 
            return constructor;
        }


        /// <summary>
        /// add assignstatement:
        /// split[1]+","+split[2]
        /// assigning corresponding split[] values
        /// </summary>
        /// <param name="member"></param>
        /// <param name="t"></param>
        /// <param name="start"></param>
        /// <param name="field"></param>
        /// <returns></returns>
        public CodeExpression cunstructConcatenation(string member, Type t, int start, int field,bool constr)
        {
            CodeExpression ass = new CodeExpression();

            if (t != typeof(int) && t != typeof(double) && t != typeof(uint) && member != "Complex" && !member.Contains("["))
            {
                // not generic
                string sep = "','";
                string concatenation = "split[" + start + "]";
                for (int j = start + 1; j < start + field; j++)
                {
                    concatenation += "+" + sep + "+" + "split[" + j + "]";
                }
                if (constr) ass = new CodeObjectCreateExpression(new CodeTypeReference(member), new CodeVariableReferenceExpression(concatenation));
                else ass = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression(member), "allowedStrings", new CodeVariableReferenceExpression(concatenation));
            }

            return ass;
        }


        public int getNumberofFields(FieldInfo[] fields, int field)
        {
            for (int i = 0; i < fields.Length; i++)
            {
                Type t = fields[i].FieldType;
                FieldInfo[] info1 = t.GetFields(BindingFlags.NonPublic | BindingFlags.Instance);
                if (t.FullName != Constants.algoB + ".FiniteLattice" && t.FullName != Constants.algoB + ".Complex")
                {
                    for (int j = 0; j < info1.Count(); j++)
                    {
                        t = info1[j].FieldType;
                        FieldInfo[] info2 = t.GetFields(BindingFlags.NonPublic | BindingFlags.Instance);
                        //under consideration of Complex and FiniteLattice!!
                        if (info2.Count() > 0 && t.FullName != Constants.algoB + ".Complex" && t.FullName != Constants.algoB + ".FiniteLattice" && t != typeof(double) && t != typeof(int) && t != typeof(uint) && !t.IsArray) field = getNumberofFields(info2, field);
                        else field += 1;
                    }
                }
                else field += 1;

                if (t.IsArray) field++;
            }

            return field;
        }

        /// <summary>
        /// Get a list with all the final datatypes, which define the elements of the corresponding semirings
        /// </summary>
        /// <param name="obj">list of datatypes</param>
        /// <param name="start">array index</param>
        /// <param name="result">list of final datatypes</param>
        public void getBaseTypeOrder(Type[] obj, int start, ref List<Type> result)
        {

            if (obj.Length > start)
            {
                // get all the private fields, which define the set of elements of the corresponding semiring!!
                if (obj[start].GetFields(BindingFlags.NonPublic | BindingFlags.Instance).Length > 0 && obj[start] != typeof(double) && obj[start] != typeof(int) && obj[start] != typeof(uint) && obj[start].Name != "Complex" && obj[start].Name != "FiniteLattice" && !obj[start].IsArray)
                {
                    // get Type array of obj[start]
                    Type[] t = new Type[obj[start].GetFields(BindingFlags.NonPublic | BindingFlags.Instance).Count()];
                    for (int i = 0; i < obj[start].GetFields(BindingFlags.NonPublic | BindingFlags.Instance).Count(); i++)
                    {
                        t[i] = obj[start].GetFields(BindingFlags.NonPublic | BindingFlags.Instance)[i].FieldType;
                    }
                    getBaseTypeOrder(t, 0, ref result);
                }
                else
                {
                    result.Add(obj[start]);
                }

                // get Type array
                getBaseTypeOrder(obj, start + 1, ref result);
            }
        }

        /// <summary>
        /// Get a list with all the final datatypes, which define the elements of the corresponding semiring-type
        /// called for more detailed generation of input-information.
        /// </summary>
        /// <param name="obj">datatyp</param>
        /// <param name="result">list of final datatypes for this type</param>
        public void getBaseTypeOrderArray(Type obj, ref List<Type> result)
        {     
                // get all the private fields, which define the set of elements of the corresponding semiring!!
                if (obj.GetFields(BindingFlags.NonPublic | BindingFlags.Instance).Length > 0 && obj!= typeof(double) && obj != typeof(int) && obj != typeof(uint) && obj.Name != "Complex" && obj.Name != "FiniteLattice")
                {
                    // get Type array of obj[start]
                    Type[] t = new Type[obj.GetFields(BindingFlags.NonPublic | BindingFlags.Instance).Count()];
                    //if(obj[start].IsArray)
            
                    for (int i = 0; i < obj.GetFields(BindingFlags.NonPublic | BindingFlags.Instance).Count(); i++)
                    {
                        t[i] = obj.GetFields(BindingFlags.NonPublic | BindingFlags.Instance)[i].FieldType;
                    }
                    getBaseTypeOrder(t, 0, ref result);
                }          
        }


        /// <summary>
        /// For the string constructor, the input for creating a Matrix<T> needs to be verified,
        /// for better overview of the CodeDOM based code, this is seperated from the String-Constructor
        /// and it also should happen before creating the single values  for a Matrix<T>. 
        /// 
        /// allowedString is based on try{converting the string} catch{ return false}, the other code is the same as for
        /// string constructor
        /// 
        /// in the case, the user is not using the automatic string-constructor generation, now method is generated
        /// </summary>
        /// <param name="attributes"></param>
        /// <returns></returns>
        public CodeMemberMethod addAllowedString(List<string> attributes, Dictionary<int, string> finite_SR)
        {
            CodeMemberMethod allowed = new CodeMemberMethod();
            allowed.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            allowed.Parameters.Add(new CodeParameterDeclarationExpression(typeof(System.String), "str"));
            allowed.Name = "allowedStrings";
            allowed.ReturnType = new CodeTypeReference("System.Boolean");

            // first statements: bool ret = true; and the specific getSplit method call, like in the string constructor
            allowed.Statements.Add(new CodeVariableDeclarationStatement(typeof(bool),"ret",new CodePrimitiveExpression(true)));

            string return_construction="";
            // generate type_info
            // get the basetype for input:
            // convert attributes into array of fieldType
            Type[] obj = new Type[attributes.Count()];
            for (int i = 0; i < attributes.Count(); i++)
            {
                Type t = getType(attributes[i]);
                obj[i] = t;
            }
            List<Type> result = new List<Type>();
            getBaseTypeOrder(obj, 0, ref result);

            // create the return array
            CodeVariableDeclarationStatement ret = new CodeVariableDeclarationStatement(typeof(string[]), "ret", new CodeArrayCreateExpression(typeof(string), result.Count));
            CodeVariableReferenceExpression retu = new CodeVariableReferenceExpression("ret");
            //allowed.Statements.Add(ret);
            // position of arrays and other basetypes
            List<int> array_info = new List<int>();
            for (int i = 0; i < result.Count(); i++)
            {
                if (result[i].IsArray) array_info.Add(i);
                if (result[i].Name == "FiniteLattice") array_info.Add(i);
            }
            CodeVariableReferenceExpression infor = new CodeVariableReferenceExpression("info");
            CodeArrayCreateExpression cai = new CodeArrayCreateExpression(typeof(int), array_info.Count());
            CodeVariableDeclarationStatement arrayinfo = new CodeVariableDeclarationStatement("System.Int32[]", "info", cai);
            allowed.Statements.Add(arrayinfo);
            for (int i = 0; i < array_info.Count(); i++)
            {
                CodeAssignStatement ass_st = new CodeAssignStatement(new CodeArrayIndexerExpression(infor, new CodePrimitiveExpression(i)), new CodePrimitiveExpression(array_info[i]));
                allowed.Statements.Add(ass_st);
            }


            allowed.Statements.Add(new CodeVariableDeclarationStatement(typeof(string[]), "split", new CodeMethodInvokeExpression(null, "getSplit", Constants.str, new CodeVariableReferenceExpression("info"), new CodePrimitiveExpression(result.Count))));
      


  
            // the dynamic generating part:
            int start = 0;
            for (int i = 0; i < attributes.Count(); i++)
            {

                //Check for MODULO and FINITELATTICE UPDATES
                if (finite_SR.Keys.Contains(i))
                {
                    string loadType = "";
                    if (attributes[i] == "Modulo") loadType = "Modulo.modulo =" + finite_SR[i];
                    if (attributes[i] == "FiniteLattice") loadType = "PreorderModel.getInstance.loadPreorder(\"" + finite_SR[i] + "\")";
                    if (attributes[i] == "LatticeBasedOnBDD") loadType = "BDDModel.getInstance.loadBDD(\"" + finite_SR[i] + "\")";
                    CodeExpressionStatement lt = new CodeExpressionStatement(new CodeVariableReferenceExpression(loadType));
                    allowed.Statements.Add(lt);
                }


               int field = 0;
               Type t = Type.GetType(attributes[i]);
               if (t == null) t = getType(attributes[i]);
               FieldInfo[] info = t.GetFields(BindingFlags.NonPublic | BindingFlags.Instance);
               if (info.Count() > 0 && attributes[i] != "Complex") field = getNumberofFields(info, field);
               else field += 1;

               CodeExpression ass = cunstructConcatenation(attributes[i], t, start, field,false);
               CodeStatement st;
               if (t == typeof(uint))
               {
                   ass = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToUInt32", new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)));
               }

               if (t == typeof(int))
               {
                   ass = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToInt32", new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)));
               }
               if (t == typeof(double))
               {
                   ass = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToDouble", new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)));
               }
               if (t.Name == getType("Complex").Name)
               {
                   if (return_construction == "") return_construction += "ret" + i.ToString();
                   else return_construction += " && ret" + i.ToString();
                   ass = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Complex"), "allowedStrings", new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)));
               }

               if (t != typeof(uint) && t != typeof(int) && t != typeof(double) && !t.IsArray)
               {
                   if(  return_construction=="")return_construction += "ret" + i.ToString();
                   else return_construction += " && ret" + i.ToString();
                   CodeVariableReferenceExpression ret_i = new CodeVariableReferenceExpression("bool ret" + i);
                   st = new CodeAssignStatement(ret_i,ass);
               }
               else
               {
                   // Defines a try statement with catch: return false
                   CodeTryCatchFinallyStatement try1 = new CodeTryCatchFinallyStatement();
                   try1.TryStatements.Add(ass);
                   CodeCatchClause catch1 = new CodeCatchClause("ex");
                   catch1.Statements.Add(new CodeCommentStatement("Handle any System.ApplicationException here."));
                   catch1.Statements.Add(new CodeMethodReturnStatement(Constants.false_expr));
                   try1.CatchClauses.Add(catch1);
                   st = try1;
               }
                                   


               if (basetypes[i].IsArray || basetypes[i].ToString().Contains("FiniteLattice"))
               {
                   if (!attributes[i].Contains("Double") && !attributes[i].Contains("Int") && !attributes[i].Contains("UInt"))
                   {
                       if (attributes[i] == "FiniteLattice")
                       {
                          // sr_type = "FiniteLattice";
                           st = new CodeExpressionStatement(new CodeVariableReferenceExpression("bool ret" + i + "=FiniteLattice.allowedStrings(split[" + i + "].Substring(1,split[" + i + "].Length-2))"));
                       }
                       else
                       {
                           //call the getSplitArrayInput method to get a List<string> values
                           CodeVariableDeclarationStatement values = new CodeVariableDeclarationStatement("List<string>", "values" + i, new CodeObjectCreateExpression("List<string>"));
                           //getArraySplit
                           allowed.Statements.Add(values);
                           CodeBinaryOperatorExpression substring_len = new CodeBinaryOperatorExpression(new CodeFieldReferenceExpression(new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)), "Length"), CodeBinaryOperatorType.Subtract, new CodePrimitiveExpression(2));
                           CodeMethodInvokeExpression substring = new CodeMethodInvokeExpression(new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)), "Substring", Constants.one, substring_len);
                           allowed.Statements.Add(new CodeMethodInvokeExpression(null, "getArraySplit", substring, new CodeVariableReferenceExpression("ref values" + i)));
                           // inside for loop, < values.Length
                           string concat = "values" + i + "[i]";
                           int finaltypes = getNumberofFields(getType(attributes[i].Substring(0, attributes[i].Length - 2)).GetFields(BindingFlags.NonPublic | BindingFlags.Instance), 0);

                           for (int j = 1; j < finaltypes; j++)
                           {
                               // not generic, a code snippet
                               concat += "+" + "','" + "+" + "values" + i + "[i+" + j + "]";

                           }


                           CodeMethodInvokeExpression tryinvoke = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression(attributes[i].Substring(0, attributes[i].Length - 2)), "allowedStrings", new CodeVariableReferenceExpression(concat));

                           // Defines a try statement with catch: return false
                           CodeTryCatchFinallyStatement try2 = new CodeTryCatchFinallyStatement();
                           try2.TryStatements.Add(tryinvoke);
                           CodeCatchClause catch2 = new CodeCatchClause("ex");
                           catch2.Statements.Add(new CodeCommentStatement("Handle any System.ApplicationException here."));
                           catch2.Statements.Add(new CodeMethodReturnStatement(Constants.false_expr));
                           try2.CatchClauses.Add(catch2);

                           CodeAssignStatement inside2 = new CodeAssignStatement(new CodeVariableReferenceExpression("i"), new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression("i"), CodeBinaryOperatorType.Add, new CodePrimitiveExpression(finaltypes-1)));
                           CodeStatement[] inside = new CodeStatement[2];
                           inside[0] = try2;
                           inside[1] = inside2;
                           st = addForLoopStatements(new CodeFieldReferenceExpression(new CodeVariableReferenceExpression("values" + i), "Count"), "i", new CodePrimitiveExpression(0), inside);
                   
                       }

                   }
                   else
                   {
                       CodeFieldReferenceExpression length = new CodeFieldReferenceExpression(new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)), "Length");
                       CodeVariableDeclarationStatement array = new CodeVariableDeclarationStatement(typeof(string), "split_array" + i, new CodeMethodInvokeExpression(new CodeArrayIndexerExpression(Constants.split, new CodePrimitiveExpression(start)), "Substring", new CodeExpression[] { new CodePrimitiveExpression(1), new CodeBinaryOperatorExpression(length, CodeBinaryOperatorType.Subtract, new CodePrimitiveExpression(2)) }));
                       allowed.Statements.Add(array);
                       // [a,b,c] -> a,b,c
                       CodeVariableDeclarationStatement array_split = new CodeVariableDeclarationStatement(typeof(string[]), "values" + i, new CodeMethodInvokeExpression(new CodeVariableReferenceExpression("split_array" + i), "Split", new CodePrimitiveExpression(',')));
                       allowed.Statements.Add(array_split);
                       CodeMethodInvokeExpression inside = new CodeMethodInvokeExpression();
                       if (attributes[i].Contains("Double"))
                       {
                           inside = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToDouble", new CodeArrayIndexerExpression(new CodeVariableReferenceExpression("values" + i), new CodeVariableReferenceExpression("i")));     
                       }
                       if (attributes[i].Contains("Int32"))
                       {
                           inside = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToInt32", new CodeArrayIndexerExpression(new CodeVariableReferenceExpression("values" + i), new CodeVariableReferenceExpression("i")));
                       }

                       if (attributes[i].Contains("UInt32"))
                       {
                           inside = new CodeMethodInvokeExpression(new CodeTypeReferenceExpression("Convert"), "ToUInt32", new CodeArrayIndexerExpression(new CodeVariableReferenceExpression("values" + i), new CodeVariableReferenceExpression("i")));
                       }

                       // Defines a try statement with catch: return false
                       CodeTryCatchFinallyStatement try3 = new CodeTryCatchFinallyStatement();
                       try3.TryStatements.Add(inside);
                       CodeCatchClause catch3= new CodeCatchClause("ex");
                       catch3.Statements.Add(new CodeCommentStatement("Handle any Exeption here."));
                       catch3.Statements.Add(new CodeMethodReturnStatement(Constants.false_expr));
                       try3.CatchClauses.Add(catch3);

                       st = addForLoop(new CodeFieldReferenceExpression(new CodeVariableReferenceExpression("values" + i), "Length"), "i", try3);
                   }
               }



               start += field;
               allowed.Statements.Add(st);
            }// end for attributes

            // return ret_i && ret_j
            if (return_construction != "")
            {
                allowed.Statements.Add(new CodeMethodReturnStatement(new CodeVariableReferenceExpression(return_construction)));
            }
            else allowed.Statements.Add(new CodeMethodReturnStatement(Constants.ret));

            return allowed;
        }

        /// <summary>
        /// Using the basetype info of the new Semiring an input information for the matrix input 
        /// is generated
        /// </summary>
        /// <param name="basetypes"></param>
        public void createInputInfo(List<Type> basetypes, Dictionary<int, string> finite_SR=null)
        {
            CodeMemberMethod temp = new CodeMemberMethod();

            temp.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            temp.Name = "inputSchema";
            temp.ReturnType = new CodeTypeReference("System.String");
            string ret ="";
            for (int i = 0; i < basetypes.Count(); i++)
            {
                if (basetypes[i].Name == "FiniteLattice")
                {
                    ret += "["+basetypes[i] + "-" + finite_SR[i]+"]";
                }
                if (basetypes[i].Name.Contains("Polynom"))
                {
                    ret += "[Polynom]";
                }
                // construct more detailed inofrmation for the user
                if (basetypes[i].IsArray)
                {
                    ret += "[" + basetypes[i]+":";
                    if (basetypes[i] != typeof(double[]) && basetypes[i] != typeof(int[]) && basetypes[i] != typeof(uint[]))
                    {
                    List<Type> typelist = new List<Type>();
                    Type t = getType(basetypes[i].Name.Substring(0, basetypes[i].Name.Length-2));
                    getBaseTypeOrderArray(t, ref typelist);
                    for (int j = 0; j < typelist.Count(); j++)
                    {
                        ret += typelist[j];
                        if (j != typelist.Count() - 1) ret += ",";
                    }

                    }
         
                    ret += "]";
                }
                else if(basetypes[i].Name!="FiniteLattice") ret += basetypes[i];
                if (i != basetypes.Count() - 1) ret += ",";
            }
            ret += " next next\nPress Enter for next row";
            // this information is to long and not really needed
            ret = ret.Replace("System.", "");
            ret = ret.Replace(Constants.algoB+".","");
            temp.Statements.Add(new CodeMethodReturnStatement(new CodePrimitiveExpression(ret)));

            targetClass.Members.Add(temp);

        }

   

        /// <summary>
        /// This method is switching between automatic generated code or codefragments given directly by the user
        /// only for methods like +,*... or ==, which are necessary for this new semiring
        /// 
        /// The behaviour of the operator is defined by his name, so the operation could be
        /// automatic implemented componentwise( if supported by attribute-TYPE) otherwise the user should specify the operation,
        /// switching between this is handled in extractMethodsStatments
        /// 
        /// For structure:s like arrays, the automatic code corresponds to convolution for *, but
        /// for / no template is provided here. This should be implemented by the user.
        /// </summary>
        /// <param name="op"></param>
        /// <param name="member"></param>
        /// <returns></returns>
        /// 
        /// code is build here on the codestatement output produced by a template or transforming string directly into c# code
        /// 
        /// </summary>
        /// <param name="input"></param>
        /// <param name="op"></param>
        /// <param name="member"></param>
        /// <returns></returns>
        public CodeStatement[] extractMethodsStatements(string input, string op, List<string> attributes, Dictionary<int, string> finite_SR)
        {
            // how many array based datatypes:
            int vector = 0;
            for (int i = 0; i < attributes.Count; i++)
            {
                if (basetypes[i].IsArray) vector++;
            }
            switch (input)
            {
                case "Componentwise-Member":
                    if (op == "==") return generateEqualityComponentwise(attributes,finite_SR);
                    if (op == "+") return generateAdditionComponentwise(attributes, vector, true, finite_SR);
                    if (op == "*") return generateMultiplicationComponentwise(attributes, vector, finite_SR);
                    // not necessary binary operations:
                    if (op == "-") return generateAdditionComponentwise(attributes, vector, false, finite_SR);
                    if (op == ">") return generateGreaterComponentwise(attributes, finite_SR);
                    return null;

                default:
                    CodeSnippetStatement[] method4 = new CodeSnippetStatement[1];
                    method4[0] = extractCode(input);
                    return method4;

            }
        }


        /// <summary>
        /// This template is for constructing the function attributes and adding the implementatin od the body by the
        /// generated statements ( done by the methods: extractMethods and generateComponentwise-methods)
        /// </summary>
        /// <param name="targetClass"></param>
        /// <param name="comment"></param>
        /// <param name="name"></param>
        /// <param name="returntype"></param>
        /// <param name="expressions"></param>
        /// <returns></returns>
        public static CodeMemberMethod addNecessaryOperators(string targetClass, string comment, string name, string returntype, CodeStatement[] expressions)
        {
            CodeMemberMethod temp = new CodeMemberMethod();
            temp.Comments.Add(new CodeCommentStatement(
            comment));
   
            temp.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            temp.Name = name;
            temp.ReturnType = new CodeTypeReference(returntype);
            if (name == "ToString")
            {

                temp.Attributes = MemberAttributes.Public | MemberAttributes.Override;
                temp.Name = name;
                temp.ReturnType = new CodeTypeReference(typeof(System.String));

            }
            if (name != "Zero" && name != "One" && name!="ToString")
            {
                if (name == "equals")
                {
                    temp.Parameters.Add(new CodeParameterDeclarationExpression(targetClass + "[]", "a"));
                    temp.Parameters.Add(new CodeParameterDeclarationExpression(targetClass + "[]", "b"));
                }
                else
                {
                    if (name == "LinearCombination")
                    {
                        temp.Parameters.Add(new CodeParameterDeclarationExpression("List<" + targetClass + "[]>", "set"));
                        temp.Parameters.Add(new CodeParameterDeclarationExpression(targetClass + "[]", "vec"));
                    }
                    else
                    {

                        temp.Parameters.Add(new CodeParameterDeclarationExpression(
                            targetClass, "a"));
                        temp.Parameters.Add(new CodeParameterDeclarationExpression(
                          targetClass, "b"));
                    }

                }
            }

            try
            {
                // this statements are generated by extractMethods || Zero & One template (getMethod)
                for (int i = 0; i < expressions.Count(); i++)
                {
                    temp.Statements.Add(expressions[i]);
                }
            }catch
            {
                MessageBox.Show("Some code-specification is missing for: "+name);
            }

            return temp;
        }


        /// <summary>
        /// generates the == operation componentwise, this means all components will be
        /// checked seperatly if they are ==.
        /// </summary>
        /// <param name="attributes"></param>
        /// <returns></returns>
        public CodeStatement[] generateEqualityComponentwise(List<string> attributes, Dictionary<int, string> finite_SR)
        {
            CodeStatement[] statement = new CodeStatement[attributes.Count + 1 + finite_SR.Count];
            int update = 0;
            string ret = "";
            for (int i = 0; i < attributes.Count; i++)
            {
                //check for updates of Modulo and FiniteLattice
                if (finite_SR.Keys.Contains(i))
                {
                    string loadType = "";
                    if (attributes[i] == "Modulo") loadType = "Modulo.modulo =" + finite_SR[i];
                    if (attributes[i] == "FiniteLattice") loadType = "PreorderModel.getInstance.loadPreorder(\"" + finite_SR[i] + "\")";
                    if (attributes[i] == "LatticeBasedOnBDD") loadType = "BDDModel.getInstance.loadBDD(\"" + finite_SR[i] + "\")";
                    CodeExpressionStatement lt = new CodeExpressionStatement(new CodeVariableReferenceExpression(loadType));
                    statement[i + update] = lt;
                    update++;
                }
                CodeFieldReferenceExpression a = new CodeFieldReferenceExpression(Constants.a,"mem"+i);
                CodeFieldReferenceExpression b = new CodeFieldReferenceExpression(Constants.b,"mem"+i);
                if (basetypes[i].IsArray)
                {
                    // call the corresponding equality method for this array type, which is also generated!!
                    statement[i+update ] = new CodeVariableDeclarationStatement (new CodeTypeReference(typeof(bool)),"ret"+i, new CodeMethodInvokeExpression(null,"equals",a,b));
                }
                else
                {
                    statement[i+update] = new CodeVariableDeclarationStatement(new CodeTypeReference(typeof(bool)), "ret" + i, new CodeBinaryOperatorExpression(a, CodeBinaryOperatorType.IdentityEquality, b));
                }
                ret += "ret" + i;
                if (i < attributes.Count - 1) ret += " && ";
            }
 

            statement[statement.Length-1] = new CodeMethodReturnStatement(new CodeVariableReferenceExpression(ret));// no language generic codegeneration
            return statement;

        }

        /// <summary>
        /// generates the + or - operation componentwise, this means all components will be
        /// checked seperatly if they are +/-.
        /// 
        /// for arrays:
        /// if b.Length> a.Length:  -b + a
        /// </summary>
        /// <param name="attributes"></param>
        /// <returns></returns>
        public CodeStatement[] generateAdditionComponentwise(List<string> attributes, int vector, bool add, Dictionary<int, string> finite_SR)
        {
            CodeBinaryOperatorType operation=  CodeBinaryOperatorType.Add;

            CodeStatement[] statement = new CodeStatement[attributes.Count - vector + 2 + (2 * vector) + finite_SR.Count];
            int update = 1;
            CodeVariableDeclarationStatement init = new CodeVariableDeclarationStatement(targetClass.Name, "ret", new CodeObjectCreateExpression(targetClass.Name));
            statement[0] = init;

            for (int i = 0; i < attributes.Count; i++)
            {
                //check for updates of Modulo and FiniteLattice
                if (finite_SR.Keys.Contains(i))
                {
                    string loadType = "";
                    if (attributes[i] == "Modulo") loadType = "Modulo.modulo =" + finite_SR[i];
                    if (attributes[i] == "FiniteLattice") loadType = "PreorderModel.getInstance.loadPreorder(\"" + finite_SR[i] + "\")";
                    if (attributes[i] == "LatticeBasedOnBDD") loadType = "BDDModel.getInstance.loadBDD(\"" + finite_SR[i] + "\")";
                    CodeExpressionStatement lt = new CodeExpressionStatement(new CodeVariableReferenceExpression(loadType));
                    statement[i + update] = lt;
                    update++;
                }
             
                CodeFieldReferenceExpression a = new CodeFieldReferenceExpression(Constants.a, "mem" + i);
                CodeFieldReferenceExpression b = new CodeFieldReferenceExpression(Constants.b, "mem" + i);
                CodeFieldReferenceExpression ret = new CodeFieldReferenceExpression(Constants.ret, "mem" + i);
                if (basetypes[i].IsArray)
                {
                    string type = attributes[i].Substring(0,attributes[i].Length-2);
                    CodeStatement[] truest = new CodeStatement[4];
                    CodeStatement[] falsest = new CodeStatement[4];
                    if (type.Contains("Double") || type.Contains("Int") || type.Contains("UInt"))
                    {
                        truest = new CodeStatement[3];
                        falsest = new CodeStatement[3];
                        truest[0] = new CodeAssignStatement(ret, new CodeArrayCreateExpression(type, new CodeFieldReferenceExpression(a, "Length")));
                        CodeAssignStatement inside_loop = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), new CodeArrayIndexerExpression(a, Constants.i));


                        truest[1] = addForLoop(new CodeFieldReferenceExpression(a, "Length"), "i", inside_loop);
                        CodeBinaryOperatorExpression op = new CodeBinaryOperatorExpression(new CodeArrayIndexerExpression(ret, Constants.i), operation, new CodeArrayIndexerExpression(b, Constants.i));
                        CodeAssignStatement inside_loop2 = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), op);
                        truest[2] = addForLoop(new CodeFieldReferenceExpression(b, "Length"), "i", inside_loop2);

                        falsest[0] = new CodeAssignStatement(ret, new CodeArrayCreateExpression(type, new CodeFieldReferenceExpression(b, "Length")));
            
                        // if op = - && b.Lenght> a.Length, then -b +a for arrays:
                        if (!add) inside_loop = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), new CodeBinaryOperatorExpression(new CodeArrayIndexerExpression(ret, Constants.i), CodeBinaryOperatorType.Subtract, new CodeArrayIndexerExpression(b, Constants.i)));
                        else inside_loop = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), new CodeArrayIndexerExpression(b, Constants.i));
                        falsest[1] = addForLoop(new CodeFieldReferenceExpression(b, "Length"), "i", inside_loop);
                        op = new CodeBinaryOperatorExpression(new CodeArrayIndexerExpression(ret, Constants.i), CodeBinaryOperatorType.Add, new CodeArrayIndexerExpression(a, Constants.i));
                        inside_loop2 = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), op);
                        falsest[2] = addForLoop(new CodeFieldReferenceExpression(a, "Length"), "i", inside_loop2);
                    }
                    else
                    {

                        //true and false state. for the ifElse

                        truest[0] = new CodeAssignStatement(ret, new CodeArrayCreateExpression(type, new CodeFieldReferenceExpression(a, "Length")));
                        CodeAssignStatement inside_loop = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), new CodeArrayIndexerExpression(a, Constants.i));
                        truest[1] = new CodeExpressionStatement(new CodeVariableReferenceExpression("for (int i = 0; i < ret.mem"+i+".Length; i++) ret.mem"+i+"[i] = " + type + ".Zero()"));

                        truest[2] = addForLoop(new CodeFieldReferenceExpression(a, "Length"), "i", inside_loop);
                        CodeBinaryOperatorExpression op = new CodeBinaryOperatorExpression(new CodeArrayIndexerExpression(ret, Constants.i), operation, new CodeArrayIndexerExpression(b, Constants.i));
                        CodeAssignStatement inside_loop2 = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), op);
                        truest[3] = addForLoop(new CodeFieldReferenceExpression(b, "Length"), "i", inside_loop2);

                        falsest[0] = new CodeAssignStatement(ret, new CodeArrayCreateExpression(type, new CodeFieldReferenceExpression(b, "Length")));
                        falsest[1] = new CodeExpressionStatement(new CodeVariableReferenceExpression("for (int i = 0; i < ret.mem" + i + ".Length; i++) ret.mem" + i + "[i] = " + type + ".Zero()"));
                        // if op = - && b.Lenght> a.Length, then -b +a for arrays:
                        if (!add) inside_loop = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), new CodeBinaryOperatorExpression(new CodeArrayIndexerExpression(ret, Constants.i), CodeBinaryOperatorType.Subtract, new CodeArrayIndexerExpression(b, Constants.i)));
                        else inside_loop = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), new CodeArrayIndexerExpression(b, Constants.i));
                        falsest[2] = addForLoop(new CodeFieldReferenceExpression(b, "Length"), "i", inside_loop);
                        op = new CodeBinaryOperatorExpression(new CodeArrayIndexerExpression(ret, Constants.i), CodeBinaryOperatorType.Add, new CodeArrayIndexerExpression(a, Constants.i));
                        inside_loop2 = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), op);
                        falsest[3] = addForLoop(new CodeFieldReferenceExpression(a, "Length"), "i", inside_loop2);

                    }


                    statement[i + update] = addIfElse(truest, falsest, new CodeFieldReferenceExpression(a, "Length"), new CodeFieldReferenceExpression(b, "Length"), CodeBinaryOperatorType.GreaterThanOrEqual);
                    update++;
                    CodeMethodInvokeExpression ifzero = new CodeMethodInvokeExpression(null, "ifZeroCut", ret);
                    statement[i + update] = new CodeAssignStatement(ret,ifzero);
                }
                else
                {
                    CodeBinaryOperatorExpression plus = new CodeBinaryOperatorExpression(a,CodeBinaryOperatorType.Add,b);
                    statement[i + update] = new CodeAssignStatement(ret,plus);
                }
            }
            statement[statement.Count() - 1] = new CodeMethodReturnStatement(Constants.ret);
            return statement;
        }


        /// <summary>
        /// generates the * operation componentwise, this means all components will be
        /// checked seperatly if they are *.
        /// </summary>
        /// <param name="attributes"></param>
        /// <returns></returns>
        public CodeStatement[] generateMultiplicationComponentwise(List<string> attributes, int vector, Dictionary<int, string> finite_SR)
        {
            int array_sr = 0;
            for (int i = 0; i < attributes.Count; i++)
            {
                if (basetypes[i].IsArray && basetypes[i] != typeof(double[]) && basetypes[i] != typeof(int[]) && basetypes[i] != typeof(uint[])) array_sr++;
            }
            CodeStatement[] statement = new CodeStatement[attributes.Count - vector + 2 + (3 * vector)+ array_sr +finite_SR.Count];
     
            int update = 1;
            CodeVariableDeclarationStatement init = new CodeVariableDeclarationStatement(classname, "ret", new CodeObjectCreateExpression(classname));
            statement[0] = init;
            for (int i = 0; i < attributes.Count; i++)
            {
                //check for updates of Modulo and FiniteLattice
                if (finite_SR.Keys.Contains(i))
                {
                    string loadType = "";
                    if (attributes[i] == "Modulo") loadType = "Modulo.modulo =" + finite_SR[i];
                    if (attributes[i] == "FiniteLattice") loadType = "PreorderModel.getInstance.loadPreorder(\"" + finite_SR[i] + "\")";
                    if (attributes[i] == "LatticeBasedOnBDD") loadType = "BDDModel.getInstance.loadBDD(\"" + finite_SR[i] + "\")";
                    CodeExpressionStatement lt = new CodeExpressionStatement(new CodeVariableReferenceExpression(loadType));
                    statement[i + update] = lt;
                    update++;
                }

                CodeFieldReferenceExpression a = new CodeFieldReferenceExpression(Constants.a, "mem" + i);
                CodeFieldReferenceExpression b = new CodeFieldReferenceExpression(Constants.b, "mem" + i);
                CodeFieldReferenceExpression ret = new CodeFieldReferenceExpression(Constants.ret, "mem" + i);
                if (basetypes[i].IsArray)
                {
                    string type = attributes[i].Substring(0, attributes[i].Length - 2);
                    CodeBinaryOperatorExpression len = new CodeBinaryOperatorExpression(new CodeFieldReferenceExpression(a, "Length"), CodeBinaryOperatorType.Add, new CodeFieldReferenceExpression(b, "Length"));
                    statement[i + update] = new CodeAssignStatement(ret,new CodeArrayCreateExpression(type,len));
                    update++;
                    //update init array, if semiring
                    if (basetypes[i] != typeof(double[]) && basetypes[i] != typeof(int[]) && basetypes[i] != typeof(uint[]))
                    {
                        statement[i + update] = new CodeExpressionStatement(new CodeVariableReferenceExpression("for (int i = 0; i < ret.mem"+i+".Length; i++) ret.mem"+i+"[i] = "+type+".Zero()"));
                        update++;
                    }
                    CodeBinaryOperatorExpression multilpy = new CodeBinaryOperatorExpression(new CodeArrayIndexerExpression(a,Constants.j),CodeBinaryOperatorType.Multiply, new CodeArrayIndexerExpression(b,Constants.i));
                    CodeBinaryOperatorExpression index = new CodeBinaryOperatorExpression(Constants.i, CodeBinaryOperatorType.Add, Constants.j);
                    CodeBinaryOperatorExpression add = new CodeBinaryOperatorExpression(new CodeArrayIndexerExpression(ret, index),CodeBinaryOperatorType.Add ,multilpy);
                    CodeStatement innerloop = addForLoop(new CodeFieldReferenceExpression(b, "Length"), "i", new CodeAssignStatement(new CodeArrayIndexerExpression(ret, index), add));
                    statement[i + update] = addForLoop(new CodeFieldReferenceExpression(a, "Length"), "j", innerloop, CodeBinaryOperatorType.Add);
                    
                    update++;
                    CodeMethodInvokeExpression ifzero = new CodeMethodInvokeExpression(null, "ifZeroCut", ret);
                    statement[i + update] = new CodeAssignStatement(ret, ifzero);
                }
                else
                {
                    CodeBinaryOperatorExpression mul = new CodeBinaryOperatorExpression(a, CodeBinaryOperatorType.Multiply, b);
                    statement[i + update] = new CodeAssignStatement(ret, mul);
                }
            }
            statement[statement.Count() - 1] = new CodeMethodReturnStatement(Constants.ret);
            return statement;
        }


        /// <summary>
        /// generates the > operation componentwise, this means all components will be
        /// checked seperatly if they are >.
        /// </summary>
        /// <param name="attributes"></param>
        /// <returns></returns>
        public CodeStatement[] generateGreaterComponentwise(List<string> attributes, Dictionary<int, string> finite_SR)
        {
            CodeStatement[] statement = new CodeStatement[attributes.Count + 2 + finite_SR.Count];
            int update = 1;
            statement[0] = new CodeVariableDeclarationStatement(typeof(bool), "ret", Constants.true_expr);
            for (int i = 0; i < attributes.Count; i++)
            {
                //check for updates of Modulo and FiniteLattice
                if (finite_SR.Keys.Contains(i))
                {
                    string loadType = "";
                    if (attributes[i] == "Modulo") loadType = "Modulo.modulo =" + finite_SR[i];
                    if (attributes[i] == "FiniteLattice") loadType = "PreorderModel.getInstance.loadPreorder(\"" + finite_SR[i] + "\")";
                    if (attributes[i] == "LatticeBasedOnBDD") loadType = "BDDModel.getInstance.loadBDD(\"" + finite_SR[i] + "\")";
                    CodeExpressionStatement lt = new CodeExpressionStatement(new CodeVariableReferenceExpression(loadType));
                    statement[i + update] = lt;
                    update++;
                }

                CodeFieldReferenceExpression a = new CodeFieldReferenceExpression(Constants.a, "mem" + i);
                CodeFieldReferenceExpression b = new CodeFieldReferenceExpression(Constants.b, "mem" + i);

                if (basetypes[i].IsArray)
                {
                    // truestatement part
                    CodeStatement truest_inner1 = new CodeMethodReturnStatement(Constants.false_expr);
                    CodeFieldReferenceExpression aindex = new CodeFieldReferenceExpression(a,"Length");
                    CodeFieldReferenceExpression bindex = new CodeFieldReferenceExpression(b, "Length");
                    CodeStatement if_inner1 = addIF(truest_inner1, new CodeArrayIndexerExpression(a, Constants.j), new CodeArrayIndexerExpression(b, Constants.j), CodeBinaryOperatorType.LessThan);                    
                    CodeStatement true_for = addForLoop(aindex, Constants.zero, "j", if_inner1, CodeBinaryOperatorType.Subtract);                  
                    // falsestatement part
                    CodeStatement ifelse = addIfElse(new CodeCommentStatement("No return for true, depends on all components!"), new CodeMethodReturnStatement(Constants.false_expr), aindex, bindex, CodeBinaryOperatorType.GreaterThan);                           
                    statement[i + update] = addIfElse(true_for, ifelse, aindex, bindex, CodeBinaryOperatorType.IdentityEquality);
                }
                else
                {  // the attribut i should be really bigger
                    statement[i +update] = new CodeAssignStatement(Constants.ret, new CodeBinaryOperatorExpression(a,CodeBinaryOperatorType.GreaterThan,b));
                }
            }
            statement[attributes.Count + 1 + finite_SR.Count] = new CodeMethodReturnStatement(Constants.ret);
            return statement;
        }

        /// <summary>
        /// automatic Zero() and One() componentwise generating template
        /// </summary>
        /// <param name="member"></param>
        /// <param name="method"></param>
        /// <returns></returns>
        public CodeStatement[] getMethod(List<string> member, string method, Dictionary<int, string> finite_SR)
        {

            int array = 0;
   
                for (int i = 0; i < member.Count; i++)
                {
                    if (basetypes[i].IsArray) array++;
                }

            CodeExpression[] parameters = new CodeExpression[member.Count];
            CodeStatement[] st = new CodeStatement[member.Count()+array + 1 +finite_SR.Count()];
            int a = 0;
            for (int i = 0; i < member.Count(); i++)
            {

                if (finite_SR.Keys.Contains(i))
                {
                    string loadType = "";
                    if (member[i] == "Modulo") loadType = "Modulo.modulo =" + finite_SR[i];
                    if (member[i] == "FiniteLattice") loadType = "PreorderModel.getInstance.loadPreorder(\"" + finite_SR[i] + "\")";
                    if (member[i] == "LatticeBasedOnBDD") loadType = "BDDModel.getInstance.loadBDD(\"" + finite_SR[i] + "\")";
                    CodeExpressionStatement lt = new CodeExpressionStatement(new CodeVariableReferenceExpression(loadType));
                    st[a] = lt;
                    a++;
                }


                CodeVariableReferenceExpression var = new CodeVariableReferenceExpression("x"+i);
                parameters[i] = var;
                CodeExpression create = new CodeMethodInvokeExpression(new CodeVariableReferenceExpression(member[i]),method.Substring(0,method.Length-2));
                if (basetypes[i].IsArray) create = new CodeArrayCreateExpression(new CodeTypeReference(basetypes[i]), 1);
                if (member[i] == "System.Double" || member[i] == "System.Int32" || member[i] == "System.UInt32")
                {
                    create = Constants.zero;
                    if (method == "One()") create = Constants.one;
                }
                st[a] = new CodeVariableDeclarationStatement(new CodeTypeReference(basetypes[i]), "x" + i, create);
                a++;
                if (basetypes[i].IsArray && method == "One()")
                {
                    CodeArrayIndexerExpression left = new CodeArrayIndexerExpression(var, Constants.zero);
                    CodeExpression right;
                    if (member[i] == "System.Double[]" || member[i] == "System.Int32[]" || member[i] == "System.UInt32[]")
                    {
                        right= Constants.one;
                    }else
                    {
                        right = new CodeMethodInvokeExpression(new CodeVariableReferenceExpression(member[i].Substring(0,member[i].Length-2)), "One");
                    }
                    st[a] = new CodeAssignStatement(left,right);
                    a++;
                }
                if (basetypes[i].IsArray && method == "Zero()")
                {
                    CodeArrayIndexerExpression left = new CodeArrayIndexerExpression(var, Constants.zero);
                    CodeExpression right;
                    if (member[i] == "System.Double[]" || member[i] == "System.Int32[]" || member[i] == "System.UInt32[]")
                    {
                        right = Constants.zero;
                    }
                    else
                    {
                        right = new CodeMethodInvokeExpression(new CodeVariableReferenceExpression(member[i].Substring(0, member[i].Length - 2)), "Zero");
                    }
                    st[a] = new CodeAssignStatement(left, right);
                    a++;
                }
            }

            CodeObjectCreateExpression ret = new CodeObjectCreateExpression(targetClass.Name,parameters);
            st[st.Length - 1] = new CodeMethodReturnStatement(ret);
            return st;
        }


        /// <summary>
        /// TOString() Template
        /// </summary>
        /// <param name="member"></param>
        /// <returns></returns>
        public CodeMemberMethod getToString(List<string> member)
        {
            CodeMemberMethod method = new CodeMemberMethod();
            method.Comments.Add(new CodeCommentStatement(
            "ToString()"));
            method.Attributes =
                MemberAttributes.Public | MemberAttributes.Override;
            method.Name = "ToString";
            method.ReturnType =
                new CodeTypeReference(typeof(System.String));

            string sep = "+\",\"";
            string sep1 = "\"/\"+";
            string ret = "return (";
            for (int i = 0; i < member.Count(); i++)
            {
                // not modeled as AST for every final expression
                // first check if idea works, than expressed as complete AST (if enough time)
                if (member[i].Contains("[]"))
                {
                    string zero = member[i].Substring(0, member[i].Length - 2)+".Zero()";
                    if (basetypes[i] == typeof(double[]) || basetypes[i] == typeof(int[]) || basetypes[i] == typeof(uint[])) zero = "0";
                    method.Statements.Add(genDeclareInitExpression(typeof(System.String), "array" + i, "\"\""));
                    CodeStatement[] insideloop = new CodeStatement[2];
                    insideloop[0] = (new CodeExpressionStatement(new CodeVariableReferenceExpression("if(i!=mem" + i + ".Length-1 && mem" + i + "[i]!="+zero+") array" + i + "+=i+\":\"+mem" + i + "[i].ToString().Replace(',','.')" + sep)));
                    insideloop[1] = (new CodeExpressionStatement(new CodeVariableReferenceExpression("else if( mem" + i + "[i]!="+zero+") array" + i + "+=i+\":\"+mem" + i + "[i].ToString().Replace(',','.')")));
                    method.Statements.Add(addForLoopStatements("mem" + i + ".Length", "i", 0, insideloop));
                    if (i != member.Count() - 1) ret += "array" + i + "+" + sep1;
                    else ret += "array" + i + ")";

                }
                else
                {
                    if (i != member.Count() - 1) ret += "mem" + i + ".ToString()" + "+" + sep1;
                    else ret += "mem" + i + ".ToString())";
                }

            }
            method.Statements.Add(new CodeExpressionStatement(new CodeVariableReferenceExpression(ret)));
            return method;

        }




        /// <summary>
        /// For the user it is possible, to generate more methods for the behaviour of his module or semiring.
        /// The code comes as string from the GUI input and is transformed into code based on extractCode( input).
        /// </summary>
        /// <param name="name"></param>
        /// <param name="code"></param>
        /// <param name="para"></param>
        /// <param name="ret"></param>
        /// <returns></returns>
        public CodeMemberMethod getOptionalMethod(string name, string code, string para, string ret)
        {

            CodeMemberMethod optional = new CodeMemberMethod();
            optional.Attributes = MemberAttributes.Private;

            optional.Name = name;

            // get type by string

            if (ret != "")
            {
                string returntype = targetClass.Name;
                if (ret == "string") returntype = "System.String";
                if (ret == "int") returntype = "System.Int32";
                if (ret == "double") returntype = "System.Double";
                if (ret == "double[]") returntype = "System.Double []";
                if (ret == "List<string>") returntype = "List<String>";
                if (ret == "List<NewClass>") returntype = "List<" + targetClass.Name + ">";
                optional.ReturnType = new CodeTypeReference(returntype);
            }

            string[] parameters = para.Split(new Char[] { ',' });
            int i = 0;
            foreach (string s in parameters)
            {
                if (!String.IsNullOrWhiteSpace(s))
                {
                    string parameter = s;
                    try
                    {
                        if (s == "string") parameter = "System.String";
                        if (s == "int") parameter = "System.Int32";
                        if (s == "double") parameter = "System.Double";
                        if (s == "double[]") parameter = "System.Double []";
                        if (s == "List<string>") parameter = "List<String>";
                        if (s == "List<NewClass>") parameter = "List<" + targetClass.Name + ">";

                        parameter = getType(parameter).ToString();
                        optional.Parameters.Add(new CodeParameterDeclarationExpression(new CodeTypeReference(parameter), "a" + i));
                    }
                    catch
                    {
                        MessageBox.Show(parameter  + " is not a type or you do not use the C#-name for the types based on System.! Your Method: " + name +" is not fully implemented.");
                        return optional;
                    }
                    i++;
                }
            }

            optional.Statements.Add(extractCode(code));

            return optional;
        }



        /// <summary>
        /// Working with arrays requires a way to compare for equality
        /// (needed for : Addition of fractions, == Operator)
        /// </summary>
        /// <param name="para1"></param>
        /// <returns></returns>
        protected static CodeMemberMethod equals(string para1)
        {
            CodeMemberMethod equals = new CodeMemberMethod();
            equals.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            equals.Name = "equals";
            equals.ReturnType = new CodeTypeReference("System.Boolean");
            equals.Parameters.Add(new CodeParameterDeclarationExpression(
                      new CodeTypeReference(para1), "a"));
            equals.Parameters.Add(new CodeParameterDeclarationExpression(
                new CodeTypeReference(para1), "b"));
            CodeFieldReferenceExpression alen = new CodeFieldReferenceExpression(Constants.a,"Length");
            CodeFieldReferenceExpression blen = new CodeFieldReferenceExpression(Constants.b, "Length");
            equals.Statements.Add(addIF(new CodeMethodReturnStatement(Constants.false_expr),alen,blen, CodeBinaryOperatorType.IdentityInequality));

            CodeArrayIndexerExpression ai = new CodeArrayIndexerExpression(Constants.a,Constants.i);
            CodeArrayIndexerExpression bi = new CodeArrayIndexerExpression(Constants.b, Constants.i);
            CodeStatement insideloop =addIF(new CodeMethodReturnStatement(Constants.false_expr),ai,bi,CodeBinaryOperatorType.IdentityInequality);
            equals.Statements.Add(addForLoop(alen,"i",insideloop));
            equals.Statements.Add(new CodeMethodReturnStatement(Constants.true_expr));
            return equals;
        }

        /// <summary>
        /// the negation operator
        /// </summary>
        /// <param name="name"></param>
        /// <param name="attr"></param>
        /// <param name="fieldex"></param>
        /// <returns></returns>
        protected static CodeMemberMethod minusT(string name, List<string> attr)
        {

            CodeMemberMethod minusT = new CodeMemberMethod();
            minusT.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            minusT.Name = "operator -";
            minusT.ReturnType = new CodeTypeReference(name);
            minusT.Parameters.Add(new CodeParameterDeclarationExpression(
                      new CodeTypeReference(name), "a"));
      
            minusT.Statements.Add(new CodeExpressionStatement(new CodeVariableReferenceExpression(name + " b= new " + name + "(a)")));
            for (int i = 0; i < attr.Count(); i++)
            {

                if (attr[i].Contains("[]")) minusT.Statements.Add(addForLoop("a.mem" + i + ".Length", "i", new CodeVariableReferenceExpression("b.mem" + i + "[i]=-a.mem" + i + "[i]")));
                    else minusT.Statements.Add(new CodeExpressionStatement(new CodeVariableReferenceExpression("b.mem" + i + "=-a.mem" + i)));

            }


            minusT.Statements.Add(new CodeExpressionStatement(new CodeVariableReferenceExpression("return b")));
            return minusT;
        }
        
        /// <summary>
        /// Working with arrays for Class-attributes requires methods for negation 
        /// (needed for automatic componentwise subtraction)
        /// </summary>
        /// 
        /// Compiler will infirm the user, if the type of the array does not support unary - operator
        /// <param name="para1"></param>
        /// <returns></returns>
        protected static CodeMemberMethod minus(string para1)
        {
            CodeMemberMethod minus = new CodeMemberMethod();
            minus.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            minus.Name = "minus";
            minus.ReturnType = new CodeTypeReference(para1 + "[]");
            minus.Parameters.Add(new CodeParameterDeclarationExpression(
                      new CodeTypeReference(para1 + "[]"), "a"));

            minus.Statements.Add(new CodeExpressionStatement(new CodeVariableReferenceExpression(para1 + "[] b= new " + para1 + "[a.Length]")));
            minus.Statements.Add(addForLoop("a.Length", "i", new CodeVariableReferenceExpression("b[i]=-a[i]")));
            minus.Statements.Add(new CodeExpressionStatement(new CodeVariableReferenceExpression("return b")));
            return minus;
        }

        /// <summary>
        /// Working with arrays requires a unique representation of 'Zero'
        /// example: [0][0][0][0] == [0] ? , for better handling such statements:
        /// for every array  : [1][0] becomes [1], just cut the Zeros of the array end only
        /// </summary>
        /// <param name="para1"></param>
        /// <returns></returns>
        protected static CodeMemberMethod ifZeroCut(string para1)
        {
            CodeMemberMethod equals = new CodeMemberMethod();
            equals.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            equals.Name = "ifZeroCut";
            equals.ReturnType = new CodeTypeReference(para1 + "[]");
            equals.Parameters.Add(new CodeParameterDeclarationExpression(
                      new CodeTypeReference(para1 + "[]"), "a"));

            CodeVariableReferenceExpression a = new CodeVariableReferenceExpression("a");
            CodeFieldReferenceExpression len = new CodeFieldReferenceExpression(a, "Length");
            equals.Statements.Add(addIF(new CodeMethodReturnStatement(a),len,Constants.one,CodeBinaryOperatorType.IdentityEquality));
            //[a.Length-1]
            CodeArrayIndexerExpression alen = new CodeArrayIndexerExpression(a, new CodeBinaryOperatorExpression(len, CodeBinaryOperatorType.Subtract, Constants.one));

            CodeVariableReferenceExpression ret = new CodeVariableReferenceExpression("ret");
            CodeStatement[] cut = new CodeStatement[3];
            CodeArrayCreateExpression ret_decl = new CodeArrayCreateExpression(para1, new CodeBinaryOperatorExpression(len, CodeBinaryOperatorType.Subtract, Constants.one));
            cut[0] = new CodeVariableDeclarationStatement(para1+"[]", "ret", ret_decl);

            CodeAssignStatement deepcopy = new CodeAssignStatement(new CodeArrayIndexerExpression(ret, Constants.i), new CodeArrayIndexerExpression(a, Constants.i));
            cut[1] = addForLoop(new CodeBinaryOperatorExpression(len, CodeBinaryOperatorType.Subtract, Constants.one), "i", deepcopy);
            CodeMethodInvokeExpression ifzero = new CodeMethodInvokeExpression(null, "ifZeroCut", ret);
            cut[2] = new CodeMethodReturnStatement(ifzero);

            CodeStatement[] truest = new CodeStatement[1];
            truest[0] = new CodeMethodReturnStatement(a);

            if (para1 == "Double" || para1 == "Int32" || para1 == "UInt32")
            {
                equals.Statements.Add(addIfElse(truest, cut, alen, Constants.zero, CodeBinaryOperatorType.IdentityInequality));
            }
            else
            {   // static call of Zero(), which must be implemented for every semiring!
                CodeMethodInvokeExpression zero = new CodeMethodInvokeExpression(new CodeVariableReferenceExpression(para1),"Zero");
                equals.Statements.Add(addIfElse(truest, cut, alen, zero, CodeBinaryOperatorType.IdentityInequality));
            }
            return equals;
        }


        /// <summary>
        /// Template for solving linear equatations for FIELDS (maybe the free implementation is a field)
        /// </summary>
        /// <param name="attributes"></param>
        /// <param name="targetClass"></param>
        public  void getGauss(List<string> attributes,CodeTypeDeclaration targetClass )
        {


            targetClass.Members.Add(addFLCMethods("LinearCombination", "", PredifCode.LinearCombination(targetClass.Name), "LinearCombination<" + targetClass.Name + ">", "List<" + targetClass.Name + "[]>", targetClass.Name + "[]", "", "set", "vec", ""));
            targetClass.Members.Add(addFLCMethods("DivideArray", "", PredifCode.DivideArray(targetClass.Name), "", targetClass.Name + "[]", targetClass.Name));
            targetClass.Members.Add(addFLCMethods("MultiplyArray", "", PredifCode.MultiplyArray(targetClass.Name), "", targetClass.Name + "[]", targetClass.Name));
            targetClass.Members.Add(addFLCMethods("MultiplyCopyArray", "", PredifCode.MultiplyCopyArray(targetClass.Name), targetClass.Name + "[]", targetClass.Name + "[]", targetClass.Name));
            targetClass.Members.Add(addFLCMethods("SubtractArray", "", PredifCode.SubtractArray(targetClass.Name), "", targetClass.Name + "[]", targetClass.Name + "[]"));
            targetClass.Members.Add(addFLCMethods("AddArray", "", PredifCode.AddArray(targetClass.Name), "", targetClass.Name + "[]", targetClass.Name + "[]"));
            targetClass.Members.Add(addFLCMethods("SubtractArray", "", PredifCode.SubtractArray1(targetClass.Name), "", targetClass.Name + "[]", targetClass.Name));
            targetClass.Members.Add(addFLCMethods("AddArray", "", PredifCode.AddArray1(targetClass.Name), "", targetClass.Name + "[]", targetClass.Name));
            targetClass.Members.Add(addFLCMethods("Transpose", "", PredifCode.Transpose(targetClass.Name), "List<" + classname + "[]> ", "List<" + targetClass.Name + "[]>", ""));

            targetClass.Members.Add(addFLCMethods("CopyArray", "", PredifCode.CopyArray(targetClass.Name, attributes), targetClass.Name + "[]", targetClass.Name + "[]", ""));
            targetClass.Members.Add(addFLCMethods("DeepArrayCopy", "", PredifCode.CopyArray(targetClass.Name, attributes), targetClass.Name + "[]", targetClass.Name + "[]", ""));
            targetClass.Members.Add(addFLCMethods("DeepRowListCopy", "", PredifCode.DeepRowListCopy(targetClass.Name, attributes), "List<" + targetClass.Name + "[]>", "List<" + targetClass.Name + "[]>", "", "", "set", "", ""));
            targetClass.Members.Add(addFLCMethods("DownwardElimination", "", PredifCode.DownwardElimination(targetClass.Name), "", "List<" + targetClass.Name + "[]>", targetClass.Name + "[]", "Permutation", "set", "vec", "p"));
            targetClass.Members.Add(addFLCMethods("UpwardElimination", "", PredifCode.UpwardElimination(targetClass.Name), "LinearCombination<" + targetClass.Name + ">", "List<" + targetClass.Name + "[]>", targetClass.Name + "[]", "", "set", "vec", ""));

        }


        /// <summary>
        /// generating 
        /// </summary>
        /// <param name="name"></param>
        /// <param name="comment"></param>
        /// <param name="code"></param>
        /// <param name="ret"></param>
        /// <param name="para1"></param>
        /// <param name="para2"></param>
        /// <returns></returns>
        public static CodeMemberMethod addFLCMethods(string name, string comment, string code, string ret, string para1, string para2)
        {
            CodeMemberMethod method = new CodeMemberMethod();
            method.Comments.Add(new CodeCommentStatement(comment));
            method.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            method.Name = name;
            if (ret != "") method.ReturnType = new CodeTypeReference(ret);
            if (para1 != "") method.Parameters.Add(new CodeParameterDeclarationExpression(
                      new CodeTypeReference(para1), "a"));
            if (para2 != "") method.Parameters.Add(new CodeParameterDeclarationExpression(
                   new CodeTypeReference(para2), "b"));


            method.Statements.Add(new CodeSnippetStatement(code));

            return method;
        }

        public static CodeMemberMethod addFLCMethods(string name, string comment, string code, string ret, string para1, string para2, string para3, string para1name, string para2name, string para3name)
        {
            CodeMemberMethod method = new CodeMemberMethod();
            method.Comments.Add(new CodeCommentStatement(comment));
            method.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            method.Name = name;
            if (ret != "") method.ReturnType = new CodeTypeReference(ret);
            if (para1 != "") method.Parameters.Add(new CodeParameterDeclarationExpression(
                      new CodeTypeReference(para1), para1name));
            if (para2 != "") method.Parameters.Add(new CodeParameterDeclarationExpression(
                   new CodeTypeReference(para2), para2name));
            if (para3 != "") method.Parameters.Add(new CodeParameterDeclarationExpression(
          new CodeTypeReference(para3), para3name));


            method.Statements.Add(new CodeSnippetStatement(code));

            return method;
        }


 /*
 * Following methods represents basic building templates of:
 * if
 * if else
 * for -loops
 * 
 * helping methods adapted to specific statements od CodeDom, the base to build elements of the diagram:
 * 
 * User Input code fragments(string) -> C# code( usind CodeSnippetStatement)
 * Declaration and Initialization statements
 * 
 */
        /// <summary>
        /// just a statement array and one condition-expression
        /// </summary>
        /// <param name="st"></param>
        /// <param name="con"></param>
        /// <returns></returns>
        public static CodeConditionStatement addIF(CodeStatement[] st, string con)
        {
            CodeExpression condition = new CodeVariableReferenceExpression(con);
            CodeConditionStatement ifStatement = new CodeConditionStatement(condition, st);
            return ifStatement;
        }
        //if( cond1 ) truestatement template
        public static CodeConditionStatement addIF(CodeStatement st, string be1a, string be1b, CodeBinaryOperatorType cbo)
        {
            CodeExpression condition = new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(be1a), cbo, new CodeVariableReferenceExpression(be1b));
            CodeConditionStatement ifStatement = new CodeConditionStatement(condition, st);
            return ifStatement;
        }

        public static CodeConditionStatement addIF(CodeStatement st, CodeExpression be1a, CodeExpression be1b, CodeBinaryOperatorType cbo)
        {
            CodeExpression condition = new CodeBinaryOperatorExpression(be1a, cbo, be1b);
            CodeConditionStatement ifStatement = new CodeConditionStatement(condition, st);
            return ifStatement;
        }

        /// <summary>
        /// if(cond1  binaryoperator cond2) truestatement - template
        /// </summary>
        /// <param name="innerState2"></param>
        /// <param name="be1a"></param>
        /// <param name="be1b"></param>
        /// <param name="be2a"></param>
        /// <param name="be2b"></param>
        /// <param name="cbo1"></param>
        /// <param name="cbo2"></param>
        /// <param name="main"></param>
        /// <returns></returns>
        public static CodeConditionStatement addIF(CodeStatement[] innerState2, string be1a, string be1b, string be2a, string be2b, CodeBinaryOperatorType cbo1, CodeBinaryOperatorType cbo2, CodeBinaryOperatorType main)
        {
            CodeExpression condition = new CodeBinaryOperatorExpression(
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(be1a), cbo1, new CodeVariableReferenceExpression(be1b)),
            main,
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(be2a), cbo2, new CodeVariableReferenceExpression(be2b)));
            CodeConditionStatement ifStatement = new CodeConditionStatement(condition, innerState2);

            return ifStatement;

        }
        public static CodeConditionStatement addIF(CodeStatement[] innerState2, CodeExpression be1a, CodeExpression be1b, CodeExpression be2a, CodeExpression be2b, CodeBinaryOperatorType cbo1, CodeBinaryOperatorType cbo2, CodeBinaryOperatorType main)
        {
            CodeExpression condition = new CodeBinaryOperatorExpression(
            new CodeBinaryOperatorExpression(be1a, cbo1, be1b),
            main,
            new CodeBinaryOperatorExpression(be2a, cbo2, be2b));
            CodeConditionStatement ifStatement = new CodeConditionStatement(condition, innerState2);

            return ifStatement;

        }




        public CodeConditionStatement addIfElse(CodeStatement innerState1, CodeStatement innerState2, CodeExpression first, CodeExpression sec, CodeBinaryOperatorType cbo)
        {
            CodeConditionStatement conditionalStatement = new CodeConditionStatement(
                // The condition to test.
           new CodeBinaryOperatorExpression(first, cbo, sec),
                // The statements to execute if the condition evaluates to true.
           new CodeStatement[] { innerState1 },
                // The statements to execute if the condition evalues to false.
           new CodeStatement[] { innerState2 });

            return conditionalStatement;

        }

        /// <summary>
        /// This template is for if-else blocks,with multiple statements
        /// </summary>
        /// <param name="innerState1"></param>
        /// <param name="innerState2"></param>
        /// <param name="first"></param>
        /// <param name="sec"></param>
        /// <param name="cbo"></param>
        /// <returns></returns>
        public static CodeConditionStatement addIfElse(CodeStatement[] truest, CodeStatement[] elsest, string first, string sec, CodeBinaryOperatorType cbo)
        {
            CodeConditionStatement conditionalStatement = new CodeConditionStatement(
           new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(first), cbo, new CodeVariableReferenceExpression(sec)),
           truest,
           elsest);

            return conditionalStatement;

        }
        public static CodeConditionStatement addIfElse(CodeStatement[] truest, CodeStatement[] elsest, CodeExpression first, CodeExpression sec, CodeBinaryOperatorType cbo)
        {
            CodeConditionStatement conditionalStatement = new CodeConditionStatement(
           new CodeBinaryOperatorExpression(first, cbo, sec),
           truest,
           elsest);

            return conditionalStatement;

        }


        /// <summary>
        /// For Loop template only for one statement inside the loop
        /// Input of statement in CodePrimitiveExpression for better overview.
        /// counter and iterator are dynamic
        /// </summary>
        /// <param name="counter"></param>
        /// <param name="ite"></param>
        /// <param name="st"></param>
        /// <returns></returns>

        public static CodeIterationStatement addForLoop(string counter, string ite, CodeExpression st)
        {
            CodeIterationStatement confor = new CodeIterationStatement(new CodeVariableDeclarationStatement(typeof(int), ite, new CodePrimitiveExpression(0)),
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(ite),
            CodeBinaryOperatorType.LessThan, new CodeVariableReferenceExpression(counter)),
            new CodeAssignStatement(new CodeVariableReferenceExpression(ite), new CodeBinaryOperatorExpression(
            new CodeVariableReferenceExpression(ite), CodeBinaryOperatorType.Add, new CodePrimitiveExpression(1))), new CodeExpressionStatement(st));

            return confor;
        }




        public static CodeIterationStatement addForLoop(CodeExpression counter, string ite, CodeStatement st)
        {
            CodeIterationStatement confor = new CodeIterationStatement(new CodeVariableDeclarationStatement(typeof(int), ite, new CodePrimitiveExpression(0)),
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(ite),
            CodeBinaryOperatorType.LessThan, counter),
            new CodeAssignStatement(new CodeVariableReferenceExpression(ite), new CodeBinaryOperatorExpression(
            new CodeVariableReferenceExpression(ite), CodeBinaryOperatorType.Add, new CodePrimitiveExpression(1))), st);

            return confor;
        }

        /// <summary>
        /// all the same like with PrimitiveExpression, just enables
        /// integration of other Statements like if-else or for-loops
        /// inside a for loop
        /// </summary>
        /// <param name="counter"></param>
        /// <param name="ite"></param>
        /// <param name="st"></param>
        /// <returns></returns>
        public static CodeIterationStatement addForLoop(string counter, string ite, CodeStatement st,CodeBinaryOperatorType op)
        {
            CodeIterationStatement confor = new CodeIterationStatement(new CodeVariableDeclarationStatement(typeof(int), ite, new CodePrimitiveExpression(0)),
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(ite),
            CodeBinaryOperatorType.LessThan, new CodeVariableReferenceExpression(counter)),
            new CodeAssignStatement(new CodeVariableReferenceExpression(ite), new CodeBinaryOperatorExpression(
            new CodeVariableReferenceExpression(ite), op, new CodePrimitiveExpression(1))), st);

            return confor;
        }


        public static CodeIterationStatement addForLoop(CodeExpression counter, string ite, CodeStatement st, CodeBinaryOperatorType op)
        {
            CodeIterationStatement confor = new CodeIterationStatement(new CodeVariableDeclarationStatement(typeof(int), ite, new CodePrimitiveExpression(0)),
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(ite),
            CodeBinaryOperatorType.LessThan, counter),
            new CodeAssignStatement(new CodeVariableReferenceExpression(ite), new CodeBinaryOperatorExpression(
            new CodeVariableReferenceExpression(ite), op, new CodePrimitiveExpression(1))), st);

            return confor;
        }

        public static CodeIterationStatement addForLoop(CodeExpression start, CodeExpression counter, string ite, CodeStatement st, CodeBinaryOperatorType op)
        {
            CodeIterationStatement confor = new CodeIterationStatement(new CodeVariableDeclarationStatement(typeof(int), ite, start),
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(ite),
            CodeBinaryOperatorType.GreaterThanOrEqual, counter),
            new CodeAssignStatement(new CodeVariableReferenceExpression(ite), new CodeBinaryOperatorExpression(
            new CodeVariableReferenceExpression(ite), op, new CodePrimitiveExpression(1))), st);

            return confor;
        }


        /// <summary>
        /// For Loop template for multiple CodeStatements
        /// the iterator, start and counter are dynamic
        /// </summary>
        /// <param name="counter"></param>
        /// <param name="i"></param>
        /// <param name="start"></param>
        /// <param name="statements"></param>
        /// <returns></returns>
        public static CodeIterationStatement addForLoopStatements(string counter, string i, int start, CodeStatement[] statements)
        {
            CodeIterationStatement confor = new CodeIterationStatement(new CodeVariableDeclarationStatement(typeof(int), i, new CodePrimitiveExpression(start)),
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(i),
            CodeBinaryOperatorType.LessThan, new CodeVariableReferenceExpression(counter)),
            new CodeAssignStatement(new CodeVariableReferenceExpression(i), new CodeBinaryOperatorExpression(
            new CodeVariableReferenceExpression(i), CodeBinaryOperatorType.Add, new CodePrimitiveExpression(1))),

            statements);

            return confor;
        }

        public static CodeIterationStatement addForLoopStatements(CodeExpression counter, string i, int start, CodeStatement[] statements)
        {
            CodeIterationStatement confor = new CodeIterationStatement(new CodeVariableDeclarationStatement(typeof(int), i, new CodePrimitiveExpression(start)),
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(i),
            CodeBinaryOperatorType.LessThan, counter),
            new CodeAssignStatement(new CodeVariableReferenceExpression(i), new CodeBinaryOperatorExpression(
            new CodeVariableReferenceExpression(i), CodeBinaryOperatorType.Add, new CodePrimitiveExpression(1))),

            statements);

            return confor;
        }

        public static CodeIterationStatement addForLoopStatements(CodeExpression counter, string i, CodeExpression start, CodeStatement[] statements)
        {
            CodeIterationStatement confor = new CodeIterationStatement(new CodeVariableDeclarationStatement(typeof(int), i, start),
            new CodeBinaryOperatorExpression(new CodeVariableReferenceExpression(i),
            CodeBinaryOperatorType.LessThan, counter),
            new CodeAssignStatement(new CodeVariableReferenceExpression(i), new CodeBinaryOperatorExpression(
            new CodeVariableReferenceExpression(i), CodeBinaryOperatorType.Add, new CodePrimitiveExpression(1))),

            statements);

            return confor;
        }


        /// <summary>
        /// User input by the GUI transformed to C# code
        /// </summary>
        /// <param name="input"></param>
        /// <returns></returns>
        public CodeSnippetStatement extractCode(string input)
        {

            string insert = "           ";
            input = input.Insert(0, insert);
            for (int i = 0; i < input.Length; i++)
            {
                if (input[i] == '\n')
                {
                    input = input.Insert(i + 1, insert);
                    i++;
                }
            }
            return new CodeSnippetStatement(input);

        }
        /// <summary>
        /// This statement generating method represents a declaration with initialization
        /// </summary>
        /// <param name="t"></param>
        /// <param name="varName"></param>
        /// <param name="init"></param>
        /// <returns></returns>
        public static CodeVariableDeclarationStatement genDeclareInitExpression(Type t,string varName,string init)
        {           
            CodeVariableDeclarationStatement variableDeclaration = new CodeVariableDeclarationStatement(
            // Type of the variable to declare.
             t,
            // Name of the variable to declare.
            varName,
           // Optional initExpression parameter initializes the variable.
            new CodeVariableReferenceExpression(init));

            return variableDeclaration;
        }



        /// <summary>
        /// CodeDomProcider:
        /// This class is the basic abstract class for building instances of code generators and compilers.
        /// Code-Generator: generate code in C#:
        /// </summary>
        public void GenerateCSharpCode()
        {
            CodeDomProvider provider = CodeDomProvider.CreateProvider("CSharp");
            CodeGeneratorOptions options = new CodeGeneratorOptions();
            options.BracingStyle = "C";
            using (StreamWriter sourceWriter = new StreamWriter(this.outputFileName))
            {
                provider.GenerateCodeFromCompileUnit(
                    targetUnit, sourceWriter, options);
            }

        }

    }// class end
}// namespace end
